"""
Generators can be used to create universally quantified variables in a way
that preserves type safety.

Example:

```python

TreeNodeType = str

class AncestorOf(BaseModel, Fact):
    ancestor: TreeNodeType
    descendant: TreeNodeType

@axiom
def ancestor_transitivity_axiom() -> bool:
    '''For all x,y,z, if AncestorOf(x,z) and AncestorOf(z,y), then AncestorOf(x,y)'''
    return all(
        AncestorOf(ancestor=x, descendant=y)
        for x, y, z in gen3(TreeNodeType, TreeNodeType, TreeNodeType)
        if AncestorOf(ancestor=x, descendant=z) and AncestorOf(ancestor=z, descendant=y)
    )
```

The above axiom defines a universally quantified statement over the variables x, y, and z, whose
range is all strings.

Note that while the semantics of the above program are consistent with Python semantics,
actually executing the above code would take infinite time as there are infinite strings.
Instead, the python is treated as a logical specification.
"""
from typing import TypeVar, Type, Any, Generator, Tuple

T = TypeVar('T')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')


def gen(*types: Type[Any]) -> Generator[Tuple[Any, ...], None, None]:
    """
    A generator that yields tuples of values of the specified types.

    Note: this is more weakly typed than the arity-specific :ref:`gen1`, :ref:`gen2`, and :ref:`gen3` functions.

    .. note:: These are generally used in defining axioms using python syntax, rather than executed directly.

    :param types:
    :return:
    """
    while True:
        yield tuple(t() for t in types)  # Replace with actual logic


def gen1(type1: Type[T1]) -> Generator[T1, None, None]:
    """
    A generator that yields individual values of the specified type.

    .. note:: This is generally used in defining axioms using python syntax, rather than executed directly.

    :param type1:
    :return:
    """
    while True:
        yield type1()  # R


def gen2(type1: Type[T1], type2: Type[T2]) -> Generator[Tuple[T1, T2], None, None]:
    """
    A generator that yields arity 2 tuples of values of the specified types.

    .. note:: This is generally used in defining axioms using python syntax, rather than executed directly.

    :param type1:
    :param type2:
    :return:
    """
    while True:
        yield type1(), type2()  # Replace with actual logic


def gen3(type1: Type[T1], type2: Type[T2], type3: Type[T3]) -> Generator[Tuple[T1, T2, T3], None, None]:
    """
    A generator that yields arity 3 tuples of values of the specified types.

    .. note:: This is generally used in defining axioms using python syntax, rather than executed directly.

    :param type1:
    :param type2:
    :param type3:
    :return:
    """
    while True:
        yield type1(), type2(), type3()  # Replace with actual logic



from dataclasses import dataclass
from typing import ClassVar, Optional, Union

from typedlogic import Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.datamodel import NotInProfileError
from typedlogic.transformations import PrologConfig, as_fol, as_prolog


@dataclass
class PrologCompiler(Compiler):
    """A Compiler that generates Prolog code from a Theory object."""

    default_suffix: ClassVar[str] = "pro"
    config: Optional[PrologConfig] = None

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        """
        Compile a Theory object into Prolog code.

        Example:
        -------
            >>> from typedlogic import *
            >>> theory = Theory()
            >>> x = Variable("x")
            >>> theory.predicate_definitions = [PredicateDefinition("P", {"x": "str"}),
            ...              PredicateDefinition("Q", {"x": "str"})]
            >>> s = Implies(Term("P", x), Term("Q", x))
            >>> theory.add(sentence=s)
            >>> compiler = PrologCompiler()
            >>> print(compiler.compile(theory))
            %% Predicate Definitions
            % P(x: str)
            % Q(x: str)
            <BLANKLINE>
            %% Sentences
            <BLANKLINE>
            q(X) :- p(X).

        :param theory:
        :param syntax:
        :param kwargs:
        :return:

        """
        config = self.config
        if self.config is None:
            config = PrologConfig(**kwargs)
        lines = []
        lines.append("%% Predicate Definitions")
        for pd in theory.predicate_definitions:
            args = ", ".join([f"{k}: {v}" for k, v in pd.arguments.items()])
            lines.append(f"% {pd.predicate}({args})")
        for sg in theory.sentence_groups:
            lines.append(f"\n%% {sg.name}\n")
            for s in sg.sentences or []:
                try:
                    lines.append(as_prolog(s, config, translate=True))
                except NotInProfileError as e:
                    self._add_untranslatable(s)
                    fol = as_fol(s)
                    fol = fol.replace("\n", " ")
                    lines.append(f"%% UNTRANSLATABLE: {fol}")
                    if self.strict:
                        raise e
        return "\n".join(lines)



from typedlogic.integrations.frameworks.owldl.owltop import (
    IRI,
    Class,
    DataProperty,
    DataPropertyDomain,
    DataPropertyRange,
    DisjointClasses,
    EquivalentClasses,
    EquivalentObjectProperties,
    InverseObjectProperties,
    ObjectIntersectionOf,
    ObjectOneOf,
    ObjectProperty,
    ObjectPropertyDomain,
    ObjectPropertyRange,
    ObjectSomeValuesFrom,
    ObjectUnionOf,
    PropertyExpressionChain,
    SubClassOf,
    SubObjectPropertyOf,
    Thing,
    TopDataProperty,
    TopObjectProperty,
)

__all__ = [
    "IRI",
    "Thing",
    "TopObjectProperty",
    "TopDataProperty",
    "ObjectProperty",
    "DataProperty",
    "Class",
    "ObjectOneOf",
    "ObjectSomeValuesFrom",
    "ObjectIntersectionOf",
    "ObjectUnionOf",
    "SubClassOf",
    "EquivalentClasses",
    "DisjointClasses",
    "SubObjectPropertyOf",
    "EquivalentObjectProperties",
    "DataPropertyDomain",
    "DataPropertyRange",
    "ObjectPropertyDomain",
    "ObjectPropertyRange",
    "InverseObjectProperties",
    "PropertyExpressionChain",





]




from pydantic import BaseModel

from typedlogic import Fact


class FactBaseModel(BaseModel, Fact):
    """
    A Pydantic BaseModel that mixes in typed-logic Fact
    """

    def __init__(self, *args, **kwargs):
        if args:
            kwargs.update(zip(self.model_fields, args, strict=False))
        super().__init__(**kwargs)



from dataclasses import dataclass

from typedlogic import Fact, axiom
from typedlogic.theories.jsonlog.jsonlog import Key, ListNodeHasMember, Node, NodeID, ObjectNodeLookup


@dataclass(frozen=True)
class ParentChildRelationship(Fact):
    parent: NodeID
    child: NodeID


@axiom
def list_parent_child(p: NodeID, c: NodeID, ix: int):
    if ListNodeHasMember(p, ix, c):
        assert ParentChildRelationship(p, c)


@axiom
def dict_parent_child(p: NodeID, c: NodeID, k: Key):
    if ObjectNodeLookup(p, k, c):
        assert ParentChildRelationship(p, c)


@axiom
def node(p: NodeID, c: NodeID):
    if ParentChildRelationship(p, c):
        assert Node(c) and Node(p)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Optional, Type, Union

from typedlogic import Sentence, Theory
from typedlogic.parser import Parser


class ModelSyntax(str, Enum):
    UNKNOWN = "unknown"
    SEXPR = "sexpr"
    FUNCTIONAL = "functional"


@dataclass
class Compiler(ABC):
    """
    An engine for compiling from the internal logical model representation to an external format.

    Note: one of the main use cases for compiling a theory is to generate an input for a solver.
    For many use cases, it's possible to just use a Solver directly, and let the system
    take care of compilation to an intermediate form.

    You can use the registry `get_compiler` method to get a compiler for a particular syntax:

        >>> from typedlogic.registry import get_compiler
        >>> compiler = get_compiler("fol")

    Next we will transpile from Python to FOL syntax. We will use a `Parser` object:

        >>> from typedlogic.registry import get_parser
        >>> parser = get_parser("python")
        >>> theory = parser.parse_file("tests/theorems/animals.py")

    Now we will compile the theory to FOL syntax:

        >>> print(compiler.compile(theory))
        Person('Fred')
        Person('Jie')
        Animal('corky', 'cat')
        Animal('fido', 'dog')
        ∀[x:Thing species:Thing]. Animal(x, species) → Likes(x, 'Fred')
        ∀[x:Thing species:Thing]. Animal(x, 'cat') → Likes(x, 'Jie')
        ∀[x:Thing species:Thing]. Animal(x, 'dog') → ¬Likes('Fred', x)

    Another useful syntax is TPTP, which is accepted by many theorom provers:

        >>> compiler = get_compiler("tptp")
        >>> print(compiler.compile(theory))
        % Problem: animals
        fof(axiom1, axiom, person('Fred')).
        fof(axiom2, axiom, person('Jie')).
        fof(axiom3, axiom, animal('corky', 'cat')).
        fof(axiom4, axiom, animal('fido', 'dog')).
        fof(axiom5, axiom, ! [X, Species] : (animal(X, Species) => likes(X, 'Fred'))).
        fof(axiom6, axiom, ! [X, Species] : (animal(X, 'cat') => likes(X, 'Jie'))).
        fof(axiom7, axiom, ! [X, Species] : (animal(X, 'dog') => ~likes('Fred', X))).

    Another common syntax is Prolog syntax, and its variants. These are often used by Datalog solvers:

        >>> compiler = get_compiler("prolog")
        >>> print(compiler.compile(theory))
        %% Predicate Definitions
        % Likes(subject: str, object: str)
        % Person(name: str)
        % Animal(name: str, species: str)
        <BLANKLINE>
        %% persons
        <BLANKLINE>
        person('Fred').
        person('Jie').
        <BLANKLINE>
        %% animals
        <BLANKLINE>
        animal('corky', 'cat').
        animal('fido', 'dog').
        <BLANKLINE>
        %% animal_preferences
        <BLANKLINE>
        likes(X, 'Fred') :- animal(X, Species).
        likes(X, 'Jie') :- animal(X, 'cat').
        <BLANKLINE>


    """

    default_suffix: ClassVar[str] = "txt"
    parser_class: ClassVar[Optional[Type[Parser]]] = None
    strict: Optional[bool] = None


    @abstractmethod
    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        """
        Compile a theory into an external representation.

        :param theory:
        :param syntax:
        :param kwargs:
        :return:
        """
        pass

    def compile_sentence(self, sentence: Sentence, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        """
        Compiles an individual sentence

        :param sentence:
        :param syntax:
        :param kwargs:
        :return:
        """
        theory = Theory()
        theory.add(sentence)
        return self.compile(theory, syntax=syntax, **kwargs)

    @property
    def suffix(self) -> str:
        return self.default_suffix

    def _add_untranslatable(self, sentence: Sentence):
        pass



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Tuple


@dataclass
class Node:
    name: str
    positive_edges: Set[str] = field(default_factory=set)
    negative_edges: Set[str] = field(default_factory=set)
    index: int = -1
    lowlink: int = -1
    on_stack: bool = False


@dataclass
class Graph:
    nodes: Dict[str, Node] = field(default_factory=dict)
    stack: List[Node] = field(default_factory=list)
    index: int = 0
    sccs: List[List[str]] = field(default_factory=list)

    def add_edge(self, u: str, v: str, is_negative: bool = False) -> None:
        if u not in self.nodes:
            self.nodes[u] = Node(u)
        if v not in self.nodes:
            self.nodes[v] = Node(v)
        if is_negative:
            self.nodes[u].negative_edges.add(v)
        else:
            self.nodes[u].positive_edges.add(v)

    def tarjan(self) -> List[List[str]]:
        for node in self.nodes.values():
            if node.index == -1:
                self._strong_connect(node)
        return self.sccs

    def _strong_connect(self, v: Node) -> None:
        v.index = self.index
        v.lowlink = self.index
        self.index += 1
        self.stack.append(v)
        v.on_stack = True

        for w_name in v.positive_edges.union(v.negative_edges):
            w = self.nodes[w_name]
            if w.index == -1:
                self._strong_connect(w)
                v.lowlink = min(v.lowlink, w.lowlink)
            elif w.on_stack:
                v.lowlink = min(v.lowlink, w.index)

        if v.lowlink == v.index:
            scc = []
            while True:
                w = self.stack.pop()
                w.on_stack = False
                scc.append(w.name)
                if w == v:
                    break
            self.sccs.append(scc)

    def is_stratified(self) -> Tuple[bool, Optional[Tuple[str, str]]]:
        """
        Determine if the Datalog program represented by this graph is stratified.

        A Datalog program is stratified if there are no negative edges between
        predicates within the same strongly connected component (SCC).

        The method works as follows:
        1. Create a mapping from each node to its SCC index.
        2. For each node in the graph:
           a. Find its SCC index.
           b. Check all its negative edges.
           c. If any negative edge points to a node in the same SCC,
              the program is not stratified.

        Returns
        -------
            bool: True if the program is stratified, False otherwise.

        Time complexity: O(N + E), where N is the number of nodes and E is the
        total number of edges (positive and negative) in the graph.

        """
        scc_map = {node: scc_index for scc_index, scc in enumerate(self.sccs) for node in scc}
        for node_name, node in self.nodes.items():
            scc_index = scc_map[node_name]
            for neg_edge in node.negative_edges:
                if scc_map[neg_edge] == scc_index:
                    return False, (node_name, neg_edge)  # Negative edge within the same SCC
        return True, None


def analyze_datalog_program(rules: List[Tuple[str, List[Tuple[str, bool]]]]) -> Tuple[bool, Optional[Tuple[str, str]], List[List[str]]]:
    """
    Analyze a Datalog program for stratification using Tarjan's algorithm.

    Args:
    ----
    rules (List[Tuple[str, List[Tuple[str, bool]]]]): A list of rules, where each rule is represented
        as (head, [(body_pred1, is_negative1), (body_pred2, is_negative2), ...])

    Returns:
    -------
    Tuple[bool, List[List[str]]]: A tuple containing:
        - Boolean indicating whether the program is stratified
        - List of strongly connected components

    Example:
    -------
    >>> rules = [
    ...     ('p', [('q', False), ('r', True)]),  # p depends on q (positive) and r (negative)
    ...     ('q', [('s', False)]),               # q depends on s
    ...     ('r', [('t', False)]),               # r depends on t
    ...     ('s', [('p', False)]),               # s depends on p
    ...     ('t', [('u', False)])                # t depends on u
    ... ]
    >>> is_stratified, _, sccs = analyze_datalog_program(rules)
    >>> is_stratified
    True
    >>> sorted(map(sorted, sccs))  # SCCs might be in different order, so we sort them
    [['p', 'q', 's'], ['r'], ['t'], ['u']]

    Example 2 (Non-Stratified Program):
    >>> non_stratified_rules = [
    ...     ('p', [('q', False), ('r', True)]),  # p depends on q (positive) and r (negative)
    ...     ('q', [('s', False)]),               # q depends on s
    ...     ('r', [('p', False)]),               # r depends on p (creating a cycle with negative edge)
    ...     ('s', [('p', False)])                # s depends on p
    ... ]
    >>> is_stratified, e, sccs = analyze_datalog_program(non_stratified_rules)
    >>> is_stratified
    False
    >>> e
    ('p', 'r')

    """
    g = Graph()
    for head, body in rules:
        for pred, is_negative in body:
            g.add_edge(head, pred, is_negative)

    sccs = g.tarjan()
    is_stratified, causal_edge = g.is_stratified()

    return is_stratified, causal_edge, sccs




"""typedlogic package."""
from typedlogic.datamodel import (
                                  BooleanSentence, And, Or, Not, Implies,
                                  Forall, Exists, Term,
                                  NegationAsFailure, not_provable, Xor, Implied, Iff, ExactlyOne,
                                  Theory, Sentence,
                                  PredicateDefinition, SentenceGroup,
                                  Variable,
                                  )
from typedlogic.pybridge import FactMixin, Fact
from typedlogic.generators import gen, gen1, gen2, gen3
from typedlogic.decorators import axiom, goal

__all__ = [
    'BooleanSentence',
    'And',
    'Or',
    'Not',
    'Implies',
    'Iff',
    'Forall',
    'Exists',
    'Term',
    'NegationAsFailure',
    'not_provable',
    'Xor',
    'ExactlyOne',
    'Implied',
    'Theory',
    'Sentence',
    'Variable',
    'PredicateDefinition',
    'SentenceGroup',

    'Fact',
    'FactMixin',

    'gen',
    'gen1',
    'gen2',
    'gen3',

]




from typing import ClassVar, Optional, Union

from typedlogic import Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.datamodel import NotInProfileError
from typedlogic.transformations import (
    PrologConfig,
    as_prolog,
    force_stratification,
    replace_constants,
    to_horn_rules,
)


def _base_type(t: str) -> str:
    if t in ["int", "float"]:
        return "number"
    else:
        return "symbol"

def _type(name: str) -> str:
    return name.capitalize()

def _pred(name: str) -> str:
    # return name.lower()
    return name

def _var(name: str) -> str:
    return name.lower()

class SouffleCompiler(Compiler):

    default_suffix: ClassVar[str] = "dl"

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None,  **kwargs) -> str:
        blocks = []

        #for k, v in theory.constants.items():
        #    blocks.append(f".const {k}: {_base_type(v)}")
        tds = theory.type_definitions or {}
        for k, v in tds.items():
            if isinstance(v, list):
                base_types = [_base_type(x) for x in v if isinstance(x, str)]
                blocks.append(f".type {_type(k)} = {' | '.join(base_types)}")
            else:
                if not isinstance(v, str):
                    raise NotImplementedError(f"Only string types are supported; got: {v}")
                blocks.append(f".type {_type(k)} = {_base_type(v)}")

        if not theory.predicate_definitions:
            raise ValueError("No predicate definitions found in theory")
        for pd in theory.predicate_definitions:
            p = _pred(pd.predicate)
            def _ref_type(t):
                if t in theory.type_definitions:
                    return _type(t)
                else:
                    return _base_type(t)
            args = [f"{_var(v)}: {_ref_type(v_typ)}" for v, v_typ in pd.arguments.items()]
            blocks.append(f".decl {p}({', '.join(args)})")

        config = PrologConfig(
            use_lowercase_vars=True,
            use_uppercase_predicates=None,
            negation_symbol="!",
            double_quote_strings=True,
            operator_map={
                "eq": "=",
            },
            include_parens_for_zero_args=True,
        )

        horn_rules = []
        for s in theory.sentences + theory.ground_terms:
            s = replace_constants(s, theory.constants)
            # TODO: allow preserving existentials
            try:
                tr_sentences = to_horn_rules(s)
                horn_rules.extend(tr_sentences)
            except NotInProfileError:
                # blocks.append(f"% IGNORED: {s} // {e}")
                continue

        horn_rules = force_stratification(horn_rules)
        for rule in horn_rules:
            try:
                prolog = as_prolog(rule, config)
                if not prolog.endswith("."):
                    prolog += "."
                blocks.append(prolog)
            except NotInProfileError:
                # blocks.append(f"% IGNORED: {s} // {e}")
                continue

        return "\n".join(blocks)








import ast
import logging
import operator
from typing import Any, Callable, List, Mapping, Tuple, Union

from typedlogic import Implies, NegationAsFailure, Variable
from typedlogic.datamodel import (
    And,
    Exists,
    Forall,
    Iff,
    Implied,
    Not,
    Or,
    Sentence,
    SentenceGroup,
    Term,
)

logger = logging.getLogger(__name__)

AST_OP_TO_FUN: Mapping[str, Callable] = {
    'Add': operator.add,
    'Sub': operator.sub,
    'Mult': operator.mul,
    'Div': operator.truediv,
    'FloorDiv': operator.floordiv,
    'Mod': operator.mod,
    'Pow': operator.pow,
    'LShift': operator.lshift,
    'RShift': operator.rshift,
    'BitOr': operator.or_,
    'BitXor': operator.xor,
    'BitAnd': operator.and_,
    'MatMult': operator.matmul,
    # Comparison operators
    'Eq': operator.eq,
    'NotEq': operator.ne,
    'Lt': operator.lt,
    'LtE': operator.le,
    'Gt': operator.gt,
    'GtE': operator.ge,
    'Is': operator.is_,
    'IsNot': operator.is_not,
    'In': lambda x, y: x in y,
    'NotIn': lambda x, y: x not in y,
}



def parse_sentence(node: Union[ast.AST, List[ast.stmt]]) -> Sentence:
    """
    Parse an AST node into a Term instance.

        >>> tree = ast.parse("~(Person(name=x))")
        >>> func_def = tree.body[0]
        >>> sentence = parse_sentence(func_def)
        >>> assert isinstance(sentence, Not)
        >>> negated = sentence.operands[0]
        >>> assert isinstance(negated, Term)
        >>> negated
        Person(?x)

    :param node: The AST node to parse
    :type node: Union[ast.AST, List[ast.stmt]]
    :return: A Term instance
    """
    def tr_arg_or_kw_value(v: ast.expr) -> Any:
        if isinstance(v, ast.Constant):
            return v.value
        elif isinstance(v, ast.Name):
            return Variable(v.id)
        elif isinstance(v, (ast.BinOp, ast.UnaryOp, ast.Call)):
            return parse_sentence(v)
        else:
            raise ValueError(f"Unsupported argument type: {type(v)}")

    def parse_sentence_or_variable(v: ast.expr) -> Union[Sentence, Variable, Any]:
        if isinstance(v, ast.Name):
            return Variable(v.id)
        elif isinstance(v, ast.Constant):
            return v.value
        else:
            return parse_sentence(v)

    if isinstance(node, list):
        sentences = [parse_sentence(n) for n in node]
        if len(sentences) == 1:
            return sentences[0]
        else:
            return And(sentences)
    if isinstance(node, ast.Expr):
        return parse_sentence(node.value)
    if isinstance(node, ast.Assert):
        return parse_sentence(node.test)
    if isinstance(node, ast.BinOp):
        left = parse_sentence_or_variable(node.left)
        right = parse_sentence_or_variable(node.right)
        if isinstance(node.op, ast.RShift):
            assert isinstance(left, Sentence)
            assert isinstance(right, Sentence)
            return left >> right
        elif isinstance(node.op, ast.BitAnd):
            assert isinstance(left, Sentence)
            assert isinstance(right, Sentence)
            return left & right
        elif isinstance(node.op, ast.BitOr):
            assert isinstance(left, Sentence)
            assert isinstance(right, Sentence)
            return left | right
        elif isinstance(node.op, ast.BitXor):
            assert isinstance(left, Sentence)
            assert isinstance(right, Sentence)
            return left ^ right
        else:
            return Term(AST_OP_TO_FUN[node.op.__class__.__name__].__name__, left, right)
    elif isinstance(node, ast.BoolOp):
        operands = [parse_sentence(value) for value in node.values]
        if isinstance(node.op, ast.And):
            return And(*operands)
        elif isinstance(node.op, ast.Or):
            return Or(*operands)
        else:
            raise ValueError(f"Unsupported boolean operator: {type(node.op)}")
    elif isinstance(node, ast.UnaryOp):
        if isinstance(node.op, ast.Invert):
            return ~parse_sentence(node.operand)
        elif isinstance(node.op, ast.Not):
            return Not(parse_sentence(node.operand))
        elif isinstance(node.op, ast.USub):
            return NegationAsFailure(parse_sentence(node.operand))
        else:
            raise ValueError(f"Unsupported unary operator: {type(node.op)}")
    elif isinstance(node, ast.If):
        # if COND: BODY
        test = node.test
        body = node.body
        orelse = node.orelse
        if orelse:
            raise ValueError("Else clause is not supported")
        return parse_sentence(test) >> parse_sentence(body)
    elif isinstance(node, ast.Call) and (isinstance(node.func, ast.Name) and node.func.id in ["Implies", "Iff", "Implied"]):
        if len(node.args) != 2:
            raise ValueError(f"Unsupported number of arguments for {node.func.id}: {len(node.args)}")
        left = parse_sentence(node.args[0])
        right = parse_sentence(node.args[1])
        if node.func.id == "Implies":
            return Implies(left, right)
        elif node.func.id == "Iff":
            return Iff(left, right)
        elif node.func.id == "Implied":
            return Implied(left, right)
        else:
            raise AssertionError
    elif isinstance(node, ast.Call):
        predicate = get_func_name(node.func)
        if predicate in ["all", "any"]:
            if len(node.args) != 1:
                raise ValueError(f"Unsupported number of arguments for quantifier: {len(node.args)}")
            arg0 = node.args[0]
            if not isinstance(arg0, ast.GeneratorExp):
                raise ValueError(f"Unsupported argument type for quantifier: {type(arg0)}")
            args, sentence = parse_generator_node(arg0)
            if predicate == "all":
                return Forall(args, sentence)
            else:
                return Exists(args, sentence)

        def tr_keyword(kw: ast.keyword) -> Tuple[str, Any]:
            if kw.arg is None:
                raise ValueError("Positional arguments are not supported")
            if isinstance(kw.value, ast.Constant):
                v = kw.value.value
            elif isinstance(kw.value, ast.Name):
                v = Variable(kw.value.id)
            elif isinstance(kw.value, (ast.BinOp, ast.UnaryOp, ast.Call)):
                kw_val_ast = kw.value
                if not isinstance(kw_val_ast, ast.AST):
                    raise AssertionError
                v = parse_sentence(kw_val_ast)
            else:
                raise ValueError(f"Unsupported keyword value type: {type(kw.value)}")
            return kw.arg, v
        if node.keywords:
            # keyword-based arguments are translated to a dict
            bindings = dict([tr_keyword(kw) for kw in node.keywords])
            return Term(predicate, bindings)
        elif node.args:
            # positional arguments are translated to a list
            pos_args = [tr_arg_or_kw_value(arg) for arg in node.args]
            return Term(predicate, *pos_args)
        else:
            return Term(predicate, {})
    elif isinstance(node, ast.Compare):
        left = tr_arg_or_kw_value(node.left)
        if len(node.comparators) != 1:
            raise ValueError(f"Unsupported number of comparators: {len(node.comparators)}")
        right = tr_arg_or_kw_value(node.comparators[0])
        if len(node.ops) != 1:
            raise ValueError(f"Unsupported number of operators: {len(node.ops)}")
        op = node.ops[0]
        op_name = AST_OP_TO_FUN[op.__class__.__name__].__name__
        return Term(op_name, left, right)
    elif isinstance(node, ast.GeneratorExp):
        raise ValueError("Generator expressions are not supported outside all/only")
    else:
        raise NotImplementedError(f"Unsupported node type: {type(node)}")


def parse_generator_node(node: ast.GeneratorExp) -> Tuple[List[Variable], Sentence]:
    """
    Parse a generator expression node and return the arguments and the implied sentence.

    :param node:
    :return:
    """
    elt = node.elt
    gens = node.generators
    if len(gens) != 1:
        raise ValueError(f"Unsupported number of generators: {len(gens)}")
    gen = gens[0]
    iter_node = gen.iter
    if isinstance(iter_node, ast.Call):
        iter_func, iter_args = parse_gen_call(iter_node)
    else:
        # iter_func = ast.unparse(iter_node)
        iter_args = []
    if gen.ifs:
        if_exprs = [parse_sentence(if_node) for if_node in gen.ifs]
        if_conj = if_exprs[0]
        for if_expr in if_exprs[1:]:
            if_conj &= if_expr
    else:
        if_conj = None

    # arg_types = {}
    if isinstance(gen.target, ast.Name):
        arg_types = {gen.target.id: Variable(gen.target.id, iter_args[0])}
    elif isinstance(gen.target, ast.Tuple):
        arg_types = {}
        for i, xelt in enumerate(gen.target.elts):
            if not isinstance(xelt, ast.Name):
                raise ValueError(f"Unsupported target type: {type(elt)}")
            arg_types[xelt.id] = Variable(xelt.id, iter_args[i])
    else:
        raise ValueError(f"Unsupported target type: {type(gen.target)}")

    implied_expr = parse_sentence(elt)
    impl: Sentence
    if if_conj:
        impl = Implies(if_conj, implied_expr)
    else:
        impl = implied_expr
    return list(arg_types.values()), impl


def get_func_name(node: ast.expr) -> str:
    """
    Recursively get the function name from an AST node.
    """
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        return f"{get_func_name(node.value)}.{node.attr}"
    else:
        return ast.unparse(node)

def parse_gen_call(node: ast.Call) -> tuple[str, List[str]]:
    """
    Parse a gen() call node and return the function name and its arguments.
    """
    func_name = node.func.id if isinstance(node.func, ast.Name) else str(node.func)
    args = [ast.unparse(arg) for arg in node.args]
    return func_name, args


def parse_func(node: ast.Call) -> tuple[str, List[str]]:
    """
    Parse a gen() call node and return the function name and its arguments.
    """
    func_name = node.func.id if isinstance(node.func, ast.Name) else str(node.func)
    args = [ast.unparse(arg) for arg in node.args]
    return func_name, args


def parse_function_def_to_sentence_group(func_def: ast.FunctionDef) -> SentenceGroup:
    """
    Parse the AST of an axiom function and extract relevant information.

    :param func_def: The AST node of the axiom function
    :type func_def: ast.FunctionDef
    :return: A dataclass containing parsed information about the axiom
    :rtype: SentenceGroup

    :Example:

        >>> import ast
        >>> axiom_func = '''
        ... def all_persons_are_mortal_axiom() -> bool:
        ...     return all(
        ...         Person(name=x) >> Mortal(name=x)
        ...         for x in gen(NameType)
        ...     )
        ... '''
        >>> tree = ast.parse(axiom_func)
        >>> func_def = tree.body[0]  # Assuming the function is at the top level
        >>> sg = parse_function_def_to_sentence_group(func_def)
        >>> print(sg.name)
        all_persons_are_mortal_axiom
        >>> assert len(sg.sentences) == 1
        >>> sentence = sg.sentences[0]
        >>> assert isinstance(sentence, Forall)
        >>> print(sentence.quantifier)
        all
        >>> print(sentence.variables[0].name)
        x
        >>> print(sentence.variables[0].domain)
        NameType
        >>> imp = sentence.sentence
        >>> print(imp)
        (Person(?x) -> Mortal(?x))
        >>> assert isinstance(imp, Implies)
        >>> imp.operands[0]
        Person(?x)
        >>> imp.operands[0].predicate
        'Person'
        >>> imp.operands[0].bindings['name'].name
        'x'
    """
    sentence_collection = SentenceGroup(
        name=func_def.name,
        docstring=ast.get_docstring(func_def),
    )
    if not func_def.body:
        raise ValueError(f"No body found for axiom function: {func_def.name}")

    qvars = {}
    if func_def.args:
        args = func_def.args.args
        for arg in args:
            var_name = arg.arg
            if not arg.annotation:
                raise ValueError(f"Argument {var_name} is missing an annotation")
            arg_ann = arg.annotation
            if not isinstance(arg_ann, ast.Name):
                raise ValueError(f"Unsupported annotation type: {type(arg_ann)}")
            qvars[var_name] = Variable(var_name, arg_ann.id)

    sentences = []
    for body_node in func_def.body:
        def add_sentence(s: Sentence):
            if qvars:
                s = Forall([v for v in qvars.values()], s)
            sentences.append(s)
        if isinstance(body_node, ast.Return):
            if body_node.value:
                sentence = parse_sentence(body_node.value)
                add_sentence(sentence)
        elif isinstance(body_node, ast.Assert):
            sentence = parse_sentence(body_node.test)
            add_sentence(sentence)
        else:
            if isinstance(body_node, ast.If):
                sentence = parse_sentence(body_node)
                add_sentence(sentence)
            else:
                logger.info(f"Unsupported body type: {type(body_node)}")

    sentence_collection.sentences = sentences
    return sentence_collection


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from dataclasses import dataclass
from typing import ClassVar, Dict, Optional, Union

from mypy.checkexpr import defaultdict

from typedlogic import Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.datamodel import SentenceGroupType
from typedlogic.transformations import as_tptp


@dataclass
class TPTPCompiler(Compiler):

    default_suffix: ClassVar[str] = "tptp"

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        lines = [f"% Problem: {theory.name}"]
        grp_counts: Dict[str, int] = defaultdict(int)
        for sg in theory.sentence_groups:
            if sg.group_type == SentenceGroupType.GOAL:
                typ = "conjecture"
            else:
                typ = "axiom"
            for s in sg.sentences or []:
                t = as_tptp(s)
                grp_counts[typ] += 1
                lines.append(f"fof(axiom{grp_counts[typ]}, axiom, {t}).")
        return "\n".join(lines)



from dataclasses import dataclass

from typedlogic import Fact, axiom
from typedlogic.integrations.frameworks.linkml.meta import (
    Attribute,
    ClassDefinition,
    ClassSlot,
    ElementID,
    IsA,
    Mixin,
    SlotDefinition,
    SlotUsage,
)


@dataclass(frozen=True)
class InvalidClassSlot(Fact):
    cls: ElementID
    slot: ElementID

@dataclass(frozen=True)
class Disjoint(Fact):
    lhs: ElementID
    rhs: ElementID

@axiom
def disjoint(lhs: ElementID, rhs: ElementID, parent: ElementID):
    if Parent(lhs, parent) and Parent(rhs, parent) and lhs != rhs:
        assert Disjoint(lhs, rhs)


@axiom
def invalid_class_slot(cls: ElementID, slot: ElementID):
    """
    Closed world assumption rule for marking slots that are not allowed for a class
    :param cls:
    :param slot:
    :return:
    """
    if ClassDefinition(cls) and SlotDefinition(slot) and ~ClassSlot(cls, slot):
        assert InvalidClassSlot(cls, slot)

@dataclass(frozen=True)
class SlotUsageOrAttribute(Fact):
    cls: ElementID
    slot_expression: ElementID

@axiom
def slot_usage_or_attribute(cls: ElementID, slot_expression: ElementID):
    if SlotUsage(cls, slot_expression):
        assert Parent(cls, slot_expression)
    if Attribute(cls, slot_expression):
        assert Parent(cls, slot_expression)

@dataclass(frozen=True)
class Parent(Fact):
    """
    Maps to class.slots
    """

    cls: ElementID
    parent: ElementID


@axiom
def parent(element: ElementID, parent: ElementID):
    if IsA(element, parent):
        assert Parent(element, parent)
    if Mixin(element, parent):
        assert Parent(element, parent)

@axiom
def class_slot(cls: ElementID, slot: ElementID, parent: ElementID):
    if Parent(cls, parent) and ClassSlot(parent, slot):
        assert Parent(cls, slot)





"""
Assumed pre-processing:

- defaults:
    - if an axiom refers to the metaproperty P of a slot S, then defaults are NOT applied
        - default_range
        - multivalued
        - inlined (false)
"""

from typing import Any, Dict, Iterator, List, Union

from typedlogic import And, Exists, Forall, Implies, Or, PredicateDefinition, Sentence, Term, Variable, Xor
from typedlogic.integrations.frameworks.linkml.instance import (
    Association,
    InlinedObject,
    InstanceMemberType,
    NodeIsMultiValued,
    NodeIsSingleValued,
)
from typedlogic.integrations.frameworks.linkml.meta import (
    ClassDefinition,
    ClassSlot,
    Identifier,
    IsA,
    Mixin,
    TreeRoot,
    TypeDefinition,
)
from typedlogic.theories.jsonlog.jsonlog import NodeIsList, ObjectNodeLookup

Result = Union[Sentence, PredicateDefinition]

SchemaDict = Dict[str, Any]

def generate_from_object(obj: SchemaDict) -> Iterator[Sentence]:
    for k, v in obj.items():
        if k == "classes":
            for class_name, class_defn in v.items():
                yield from generate_class_definition(class_name, class_defn)
        elif k == "types":
            for type_name, type_defn in v.items():
                yield from generate_type_definition(type_name, type_defn)
    return

def generate_class_definition(class_name: str, class_defn: Dict) -> Iterator[Sentence]:
    """
    Implements mapping of a class definition to a set of sentences.

    Each class definition is treated as a collection of universally quantified sentences, of the form
    for all i, if i is an instance of class_name, then i satisfies the following conditions <...>

    This also yields tbox sentences for the class definition etc.

    :param class_name:
    :param class_defn:
    :return:
    """
    yield ClassDefinition(class_name)
    inst_var = Variable("I")
    conjs: List[Sentence] = []
    for k, v in class_defn.items():
        if k in ("slot_usage", "attributes"):
            for slot_name, slot_defn in v.items():
                for attr_k, attr_v in slot_defn.items():
                    if attr_k == "identifier":
                        yield Identifier(class_name, slot_name)
                for conj in conjunctions_from_slot_expression(inst_var, slot_name, slot_defn):
                    conjs.append(conj)
        if k == "slots":
            for slot_name in v:
                yield ClassSlot(class_name, slot_name)
        if k == "is_a":
            yield IsA(class_name, v)
            yield Forall([inst_var],
                         Implies(Term(InstanceMemberType.__name__, inst_var, class_name),
                                 Term(InstanceMemberType.__name__, inst_var, v)))
        if k == "mixins":
            for v1 in v:
                yield Mixin(class_name, v1)
                yield Forall([inst_var],
                             Implies(Term(InstanceMemberType.__name__, inst_var, class_name),
                                     Term(InstanceMemberType.__name__, inst_var, v1)))
        if k == "tree_root":
            yield TreeRoot(class_name)
            yield InstanceMemberType("/", class_name)
    if conjs:
        yield Forall([inst_var],
                     Implies(Term(InstanceMemberType.__name__, inst_var, class_name),
                             And(*conjs)))
    return

def generate_type_definition(type_name: str, type_defn: Dict) -> Iterator[Sentence]:
    yield TypeDefinition(type_name)
    inst_var = Variable("i")
    conjs: List[Sentence] = []
    for k, v in type_defn.items():
        if k == "typeof":
            yield IsA(type_name, v)
            yield Forall([inst_var],
                         Implies(Term(InstanceMemberType.__name__, inst_var, type_name),
                                 Term(InstanceMemberType.__name__, inst_var, v)))
        if k == "mixins":
            for v1 in v:
                yield Forall([inst_var],
                             Implies(Term(InstanceMemberType.__name__, inst_var, type_name),
                                     Term(InstanceMemberType.__name__, inst_var, v1)))
        #for conj in conjunctions_from_type_expression(inst_var, type_name, type_defn):
        #    conjs.append(conj)
    if conjs:
        yield Forall([inst_var],
                     Implies(Term(InstanceMemberType.__name__, inst_var, type_name),
                             And(*conjs)))
    return


def conjunctions_from_slot_expression(inst_var: Variable, slot_name: str, slot_expr: Dict) -> Iterator[Sentence]:
    """
    Generate a set of conjunctions from a slot expression

    >>> inst_var = Variable("I")
    >>> slot_name = "age"
    >>> slot_expr = {"required": True}
    >>> list(conjunctions_from_slot_expression(inst_var, slot_name, slot_expr))
    [InstanceSlotRequired(I, age)]

    :param inst_var:
    :param slot_name:
    :param slot_expr:
    :return:
    """
    val_var = Variable("v")
    assert val_var.name == "v"
    for k, v in slot_expr.items():
        if k == "any_of":
            or_exprs = []
            for sub_expr in v:
                or_exprs.extend(list(conjunctions_from_slot_expression(inst_var, slot_name, sub_expr)))
            yield Or(*or_exprs)
        elif k == "all_of":
            and_exprs = []
            for sub_expr in v:
                and_exprs.extend(list(conjunctions_from_slot_expression(inst_var, slot_name, sub_expr)))
            yield And(*and_exprs)
        elif k == "none_of":
            none_exprs = []
            for sub_expr in v:
                none_exprs.extend(list(conjunctions_from_slot_expression(inst_var, slot_name, sub_expr)))
            if len(none_exprs) == 1:
                yield ~none_exprs[0]
            else:
                yield ~And(*none_exprs)
        elif k == "exactly_one_of":
            or_exprs = []
            for sub_expr in v:
                or_exprs.extend(list(conjunctions_from_slot_expression(inst_var, slot_name, sub_expr)))
            yield Xor(*or_exprs)
        elif k == "required":
            if v:
                yield Exists([val_var], Term(Association.__name__, inst_var, slot_name, val_var))
        #elif k == "identifier":
        #    yield Identifier(class_name, slot_name)
        elif k == "range":
            # assumes pre-processing has decorated the range with the appropriate type
            # TODO: linkml:Any
            yield Forall([val_var], Implies(Term(Association.__name__, inst_var, slot_name, val_var),
                                            Term(InstanceMemberType.__name__, val_var, v)))
        elif k == "multivalued" and v is not None:
            if v is True:
                pred = NodeIsMultiValued.__name__
            else:
                pred = NodeIsSingleValued.__name__
            yield Forall([val_var], Implies(Term(ObjectNodeLookup.__name__, inst_var, slot_name, val_var),
                                            Term(pred, val_var)))
        elif k == "inlined_as_list" and v is True:
            yield Forall([val_var], Implies(Term(ObjectNodeLookup.__name__, inst_var, slot_name, val_var),
                                            Term(NodeIsList.__name__, val_var)))
            yield Forall([val_var], Implies(Term(ObjectNodeLookup.__name__, inst_var, slot_name, val_var),
                                            Term(InlinedObject.__name__, val_var, v)))
        elif k == "inlined" and v is True:
            yield Forall([val_var], Implies(Term(ObjectNodeLookup.__name__, inst_var, slot_name, val_var),
                                            Term(InlinedObject.__name__, val_var, v)))







"""
from typing import Generator, Tuple, Type, TypeVar, TypeVarTuple

class GenVar(Generic[T]):
    def __init__(self, typ: Type[T]):
        self.type = typ


Ts = TypeVarTuple('Ts')

def gen(*types: Tuple[Type[*Ts]]) -> Generator[Tuple[*Ts], None, None]:
    while True:
        yield tuple(t() for t in types)  # Replace with actual logic

# Usage example
for x, y, z in gen(str, int, float):
    print(x, y, z)  # x is recognized as str, y as int, z as float
"""



import inspect
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List, Optional, Set, Type, Union

from typedlogic import PredicateDefinition, Sentence, Theory
from typedlogic.integrations.frameworks.owldl import Thing, TopDataProperty, TopObjectProperty
from typedlogic.integrations.frameworks.owldl.owltop import Ontology
from typedlogic.integrations.solvers.clingo import ClingoSolver
from typedlogic.parsers.pyparser.python_parser import PythonParser, compile_python
from typedlogic.solver import Model, Solver
from typedlogic.utils.import_closure import compute_import_closure


def get_all_subclasses(cls: Type) -> Set[Type]:
    subclasses = set(cls.__subclasses__())
    return subclasses.union(s for c in subclasses for s in get_all_subclasses(c))


@dataclass
class OWLReasoner:
    """
    A reasoner for OWL-DL ontologies

    The reasoner is initialized with a theory, which can be loaded from a file or built up programmatically.

    Let's start with a simple data model. Note our data model must be heavily *normalized* to be visible
    to OWL-DL. All individuals are represented as subclasses of Thing. These can't (yet) have properties
    of their own, these must be represented as separate relationships, which are subclasses of TopObjectProperty.

    Our model will have a class for representing people, and a relationship for representing who knows who.

        >>> class Person(Thing):
        ...     pass
        >>> class Knows(TopObjectProperty):
        ...     domain = Person
        ...     range = Person
        ...     symmetric = True

    We can now create an OWL reasoner:

        >>> from typedlogic.integrations.frameworks.owldl.reasoner import OWLReasoner
        >>> r = OWLReasoner()
        >>> r.init_axioms()

    And add some facts. These are standard python objects:

        >>> r.add(Knows("p1", "p2"))

    We can now reason with the model:

        >>> assert r.coherent()
        >>> model = r.model()
        >>> for fact in model.iter_retrieve("knows"):
        ...     print(fact)
        knows(p1, p2)
        knows(p2, p1)

    Note that because we declared the relationship as symmetric, the reasoner inferred the reverse relationship.
    
    By default, the reasoner uses the Clingo solver. This can be changed by setting the `solver_class` attribute.

    """

    solver_class: Type[Solver] = ClingoSolver
    theory: Optional[Theory] = None
    solver: Optional[Solver] = None

    def set_solver_class(self, solver_class: Type[Solver]):
        self.solver_class = solver_class
        self.solver = None

    def init_from_file(self, source: Union[str, Path]):
        p = PythonParser()
        f = open(source)
        self.theory = p.parse(f)
        python_txt = open(source).read()
        module = compile_python(python_txt, name=None, package_path=str(source))
        # module = importlib.import_module(str(source))
        self.theory.source_module_name = module.__name__
        # TODO: ensure loaded
        self._axioms_from_classes()

    def init_axioms(self):
        self.theory = Theory()
        pp = PythonParser()
        import typedlogic.integrations.frameworks.owldl.owltop as owltop
        self.theory = pp.parse(Path(owltop.__file__))
        # Do not restrict to imports closure
        self.theory.source_module_name = None
        self._axioms_from_classes()


    def _axioms_from_classes(self):

        # find all subclasses of Thing
        class_classes = get_all_subclasses(Thing)
        op_classes = get_all_subclasses(TopObjectProperty)
        dp_classes = get_all_subclasses(TopDataProperty)
        ont_classes = get_all_subclasses(Ontology)
        all_classes = class_classes.union(op_classes).union(dp_classes).union(ont_classes)
        source_module_name = self.theory.source_module_name
        if source_module_name:
            import_closure = compute_import_closure(source_module_name)
            all_classes = [c for c in all_classes if c.__module__ in import_closure]
        for cls in all_classes:
            sentences = cls.to_sentences()
            for s in sentences:
                self.theory.add(s)

        for cls in all_classes:
            # Get the module where the class is defined
            module = inspect.getmodule(cls)

            if module:
                module_name = module.__name__

                # Check if __axioms__ is defined in the module
                if hasattr(module, '__axioms__'):
                    axioms = module.__axioms__
                    if not isinstance(axioms, list):
                        axioms = [axioms]
                    for axiom in axioms:
                        fol = axiom.as_fol()
                        if fol:
                            self.theory.add(fol)

    def add(self, sentence: Union[Sentence, List[Sentence]]):
        """
        Add a sentence to the reasoner

        :param sentence:
        :return:
        """
        self.solver = None  # solver's state is invalidated
        if isinstance(sentence, list):
            for s in sentence:
                if s is None:
                    raise ValueError(f"Got empty sentence in {sentence}")
                self.add(s)
            return
        if self.theory is None:
            self.theory = Theory()
        self.theory.add(sentence)

    def register(self, cls: Type):
        """
        Register a python class with the reasoner

        :param cls:
        :return:
        """
        if not self.theory:
            self.theory = Theory()
        pd_map = {pd.predicate: pd for pd in self.theory.predicate_definitions or []}
        pd = None
        for mc in [Thing, TopObjectProperty, TopDataProperty]:
            if issubclass(cls, mc):
                pd = PredicateDefinition(cls.__name__,
                                         pd_map[mc.__name__].arguments)
        if not pd:
            raise ValueError(f"Class {cls} is not a recognized OWL-DL class")
        self.theory.predicate_definitions.append(pd)
        sentences = cls.to_sentences()
        for s in sentences:
            self.add(s)

    def remove(self, sentence: Union[Sentence, List[Sentence]]):
        self.solver = None  # solver's state is invalidated
        if isinstance(sentence, list):
            for s in sentence:
                self.remove(s)
            return
        if self.theory is None:
            raise ValueError("No theory to remove from")
        self.theory.remove(sentence)

    def reason(self) -> None:
        self.solver = self.solver_class()
        if not self.theory:
            raise ValueError("No theory to reason with")
        self.solver.add_theory(self.theory)

    def model(self) -> Model:
        if not self.solver:
            self.reason()
        if not self.solver:
            raise ValueError("No solver to reason with")
        solver = self.solver
        return solver.model()

    def model_iter(self) -> Iterator[Model]:
        if not self.solver:
            self.reason()
        if not self.solver:
            raise ValueError("No solver to reason with")
        solver = self.solver
        yield from solver.models()

    def coherent(self) -> bool:
        if not self.solver:
            self.reason()
        if not self.solver:
            raise ValueError("No solver to reason with")
        solver = self.solver
        return solver.check().satisfiable is not False



from copy import copy
from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Dict, Iterator, List, Mapping, Optional, Set, Type, Union

import z3
from z3 import SortRef

import typedlogic as tlog
import typedlogic.pybridge
from typedlogic import FactMixin, Variable
from typedlogic.builtins import NUMERIC_BUILTINS
from typedlogic.datamodel import DefinedType, PredicateDefinition, Sentence, Term
from typedlogic.parsers.pyparser.python_ast_utils import logger
from typedlogic.profiles import (
    AllowsComparisonTerms,
    MixedProfile,
    MultipleModelSemantics,
    OpenWorld,
    Profile,
    SortedLogic,
    Unrestricted,
)
from typedlogic.pybridge import fact_arg_map, fact_predicate
from typedlogic.solver import Model, Solution, Solver

SORT_MAP: Mapping[str, Type[SortRef]] = {
    "str": z3.StringSort,
    "int": z3.IntSort,
    "bool": z3.BoolSort,
    "float": z3.RealSort,
}

# Return the first "M" models of formula list of formulas F
def get_models(s: z3.Solver, M: int) -> List[z3.Model]:
    # https://stackoverflow.com/questions/11867611/z3py-checking-all-solutions-for-equation
    # https://github.com/Z3Prover/z3/issues/5765
    result: List[z3.Model] = []
    s.push()
    while len(result) < M and s.check() == z3.sat:
        m = s.model()
        result.append(m)
        # Create a new constraint the blocks the current model
        block = []
        for d in m:
            # d is a declaration
            if d.arity() > 0:
                logger.warning(f"ignoring uninterpreted function {d}")
                continue
                #raise z3.Z3Exception(f"uninterpreted functions are not supported; {d}")
            # create a constant from declaration
            c = d()
            if z3.is_array(c) or c.sort().kind() == z3.Z3_UNINTERPRETED_SORT:
                raise z3.Z3Exception("arrays and uninterpreted sorts are not supported")
            block.append(c != m[d])
        s.add(z3.Or(block))
    s.pop()
    return result




@dataclass
class Z3Solver(Solver):
    """
    A solver that uses Z3.

        >>> from typedlogic.integrations.frameworks.pydantic import FactBaseModel
        >>> class AncestorOf(FactBaseModel):
        ...     ancestor: str
        ...     descendant: str
        >>> solver = Z3Solver()
        >>> solver.add_predicate_definition(PredicateDefinition(predicate="AncestorOf", arguments={'ancestor': "str", 'descendant': "str"}))
        >>> solver.add_fact(AncestorOf(ancestor='p1', descendant='p1a'))
        >>> solver.add_fact(AncestorOf(ancestor='p1a', descendant='p1aa'))
        >>> from typedlogic import SentenceGroup, PredicateDefinition
        >>> aa = SentenceGroup(name="transitivity-of-ancestor-of")
        >>> solver.add_sentence_group(aa)
        >>> soln = solver.check()

    """

    _wrapped_solver: Optional[z3.Solver] = None
    profile: ClassVar[Profile] = MixedProfile(Unrestricted(), SortedLogic(), MultipleModelSemantics(), OpenWorld(), AllowsComparisonTerms())
    max_models: int = field(default=5)

    # TODO: rename this
    predicate_map: Optional[Dict[str, z3.FuncDecl]] = None

    def __post_init__(self):
        if self._wrapped_solver is None:
            self._wrapped_solver = z3.Solver()

    @property
    def wrapped_solver(self) -> z3.Solver:
        if self._wrapped_solver is None:
            self._wrapped_solver = z3.Solver()
        return self._wrapped_solver

    def check(self) -> Solution:
        result = self.wrapped_solver.check()
        return Solution(satisfiable=result == z3.sat)

    def models(self) -> Iterator[Model]:
        results = get_models(self.wrapped_solver, self.max_models)
        if not results:
            raise ValueError("Not satisfiable")
        for wmodel in results:
            rmodel = Model(description=str(wmodel),
                           source_object=wmodel,
                           ground_terms=[],
                           )
            yield rmodel
        return

    def prove(self, sentence: Sentence) -> Optional[bool]:
        s = self.wrapped_solver
        s.push()
        s.add(z3.Not(self.translate(sentence)))
        result = s.check()
        s.pop()
        return result == z3.unsat

    def _unroll_type(self, typ: DefinedType) -> Set[str]:
        """
        Unroll a defined type into its components

        :param typ:
        :return:
        """
        if isinstance(typ, str):
            if typ in self.type_definitions:
                return self._unroll_type(self.base_theory.type_definitions[typ])
            return {typ}
        if isinstance(typ, list):
            ts: List[str] = []
            for t in typ:
                ts.extend(self._unroll_type(t))
            return set(ts)
        raise ValueError(f"Unknown type {typ}")

    def _sort(self, typ: Optional[str] = None) -> Union[Callable, Type[z3.SortRef]]:
        # TODO: change this to return instantiated sort, not the sort class
        if not typ:
            return z3.StringSort
        typs = self._unroll_type(typ)
        repl_map = {
            "Decimal": "float",
        }
        typs = {repl_map.get(t, t) for t in typs}
        if "float" in typs and "int" in typs:
            typs = typs.difference({"int"})
        if len(typs) > 1:
            # unions not directly supported
            # TODO: add constraints
            return lambda: z3.DeclareSort(typ)
        typ = list(typs)[0]
        if not isinstance(typ, str):
            # TODO - we should ensure types are strings
            typ = typ.__name__
        return SORT_MAP.get(typ, z3.StringSort)

    def _const(self, value: Any, typ: str) -> z3.Const:
        sort = self._sort(typ)
        return z3.Const(value, sort())

    def _func(self, name: str, *args) -> z3.FuncDecl:
        args = tuple([args] + [z3.BoolSort()])
        return z3.Function(name, *args)

    def _tr(self, var: Any, bindings: dict) -> z3.ExprRef:
        if var is None:
            return z3.StringVal("None")
        if isinstance(var, Variable):
            if var.name not in bindings:
                if var.name in self.constants:
                    pf_arg = self.constants[var.name]
                else:
                    raise ValueError(f"Variable {var.name} not bound in {bindings} or {self.constants}")
            else:
                pf_arg = bindings[var.name]
            return pf_arg
        py_typ = type(var).__name__
        z3_typ = self._sort(py_typ)
        t2m_map = {
            z3.StringSort: z3.StringVal,
            z3.IntSort: z3.IntVal,
            z3.BoolSort: z3.BoolVal,
            z3.RealSort: z3.RealVal,
        }
        z3_valf = t2m_map.get(z3_typ, z3.StringVal)
        return z3_valf(var)

    def add_fact(self, fact: FactMixin) -> None:
        return self.add_sentence(fact)

    def add_sentence(self, sentence: Sentence) -> None:
        # normalize_variables(sentence)
        z3_expr = self.translate(sentence)
        self.wrapped_solver.add(z3_expr)


    def add_predicate_definition(self, predicate_definition: PredicateDefinition) -> None:
        """
        Add a predicate definition to the solver.

        E.g. `` Function('AncestorOf', StringSort(), StringSort(), BoolSort())``

        :param predicate_definition:
        :return:
        """
        args = [self._sort(a)() for a in predicate_definition.arguments.values()]
        args += [z3.BoolSort()]
        p = z3.Function(predicate_definition.predicate, *args)
        if not self.predicate_map:
            self.predicate_map = {}
        self.predicate_map[predicate_definition.predicate] = p
        if not self.predicate_definitions:
            self.predicate_definitions = {}
        self.predicate_definitions[predicate_definition.predicate] = predicate_definition

    def translate(self, sentence: Sentence, bindings: Optional[Dict[str, z3.SortRef]] = None) -> z3.ExprRef:
        """
        Translate a Sentence to a Z3 expression.

        :param sentence: The Sentence to translate
        :param bindings: local bindings of variable names to Z3 Sorts
        :return: The Z3 expression
        """
        if isinstance(sentence, tlog.And):
            return z3.And(*[self.translate(op, bindings) for op in sentence.operands])
        if isinstance(sentence, tlog.Or):
            return z3.Or(*[self.translate(op, bindings) for op in sentence.operands])
        if isinstance(sentence, tlog.Xor):
            return z3.Xor(*[self.translate(op, bindings) for op in sentence.operands])
        if isinstance(sentence, tlog.ExactlyOne):
            disj = []
            for a in sentence.operands:
                disj.append(z3.And(self.translate(a, bindings),
                                   *[z3.Not(self.translate(b, bindings)) for b in sentence.operands if b != a]))
            return z3.Or(*disj)
        if isinstance(sentence, tlog.Not):
            return z3.Not(self.translate(sentence.operands[0], bindings))
        if isinstance(sentence, tlog.Iff):
            # rewrite
            lhs = sentence.left
            rhs = sentence.right
            rewritten = tlog.And(tlog.Implies(lhs, rhs), tlog.Implies(rhs, lhs))
            return self.translate(rewritten, bindings)
        if isinstance(sentence, tlog.Implied):
            # rewrite
            lhs = sentence.operands[0]
            rhs = sentence.operands[1]
            return self.translate(tlog.Implies(rhs, lhs), bindings)
        if isinstance(sentence, tlog.Implies):
            lhs = self.translate(sentence.operands[0], bindings)
            rhs = self.translate(sentence.operands[1], bindings)
            return z3.Implies(lhs, rhs)
        if isinstance(sentence, (tlog.Forall, tlog.Exists)):
            if not bindings:
                bindings = {}
            args = []
            for v in sentence.variables:
                var_name = v.name
                domain = v.domain
                s = self._sort(domain)
                arg = z3.Const(var_name, s())  ## TODO
                bindings[var_name] = arg
                args.append(arg)
            inner_sentence = self.translate(sentence.sentence, bindings)
            if isinstance(sentence, tlog.Exists):
                return z3.Exists(args, inner_sentence)
            else:
                return z3.ForAll(args, inner_sentence)
        if isinstance(sentence, (tlog.Term, typedlogic.pybridge.FactMixin)): # TODO: use Expression
            if isinstance(sentence, typedlogic.pybridge.FactMixin):
                sentence = tlog.Term(fact_predicate(sentence), fact_arg_map(sentence))
            if not self.predicate_map or not self.predicate_definitions:
                raise ValueError("You must add predicate definitions before adding facts")
            pd = self.predicate_definitions.get(sentence.predicate)
            pf = self.predicate_map.get(sentence.predicate)
            if pf is None and sentence.predicate in NUMERIC_BUILTINS:
                pf = NUMERIC_BUILTINS[sentence.predicate]
            elif pf is None or pd is None:
                raise ValueError(f"Predicate {sentence.predicate} not found in {self.predicate_map}\n"
                                 "Did you remember to declare these as predicates?")
            elif sentence.positional:
                # TODO: more elegant way to do this
                sentence = copy(sentence)
                sentence.make_keyword_indexed(list(pd.arguments.keys()))
            pf_args = []
            for arg_name, var in sentence.bindings.items():
                if not bindings:
                    bindings = {}
                if isinstance(var, Variable):
                    if var.name not in bindings:
                        if var.name in self.constants:
                            pf_arg = self.constants[var.name]
                        else:
                            raise ValueError(f"Variable {var.name} not bound in {bindings} or {self.constants}")
                    else:
                        pf_arg = bindings[var.name]
                    pf_args.append(pf_arg)
                elif isinstance(var, Term):
                    args = [self._tr(a, bindings) for a in var.values]
                    p = var.predicate
                    if p == "add":
                        pf_args.append(args[0] + args[1])
                    elif p == "gt":
                        pf_args.append(args[0] > args[1])
                    else:
                        raise NotImplementedError(f"Term not implemented: {var}")
                elif var is None:
                    pf_args.append(z3.StringVal("None"))
                else:
                    py_typ = type(var).__name__
                    z3_typ = self._sort(py_typ)
                    t2m_map = {
                        z3.StringSort: z3.StringVal,
                        z3.IntSort: z3.IntVal,
                        z3.BoolSort: z3.BoolVal,
                        z3.RealSort: z3.RealVal,
                    }
                    z3_valf = t2m_map.get(z3_typ, z3.StringVal)
                    pf_arg = z3_valf(var)
                    pf_args.append(pf_arg)
            try:
                z3_expr = pf(*pf_args)
            except Exception as e:
                raise ValueError(f"Error translating {sentence} args: {pf_args} to Z3 using {pf}:\n{e}")
            return z3_expr
        raise NotImplementedError(f'Not implemented:{type(sentence)} :: {sentence}')

    def dump(self) -> str:
        return str(self.wrapped_solver)




import importlib


def compute_import_closure(root_module_name):
    """
    Compute the import closure of a module.

    :param root_module_name:
    :return:
    """
    closure = {root_module_name}
    visited = set()

    def dfs(module_name):
        if module_name in visited:
            return
        visited.add(module_name)
        closure.add(module_name)

        module = importlib.import_module(module_name)


        # closure[root_module_name].add(module_name)

        for name, _ in module.__dict__.items():
            if name.startswith('__'):
                continue
            try:
                submodule = getattr(module, name)
                if hasattr(submodule, '__module__'):
                    dfs(submodule.__module__)
            except:
                pass

    dfs(root_module_name)
    return closure


def get_class_module(cls):
    return cls.__module__


def check_class_module_and_closure(c, closure):
    class_module = get_class_module(c)
    is_in_closure = class_module in closure

    print(f"Class {c.__name__} is declared in module: {class_module}")
    print(f"Is {class_module} in the import closure? {is_in_closure}")




"""
Data model for the typed-logic framework.

This module defines the core classes and structures used to represent logical
constructs such as sentences, terms, predicates, and theories. It is based
on the Common Logic Interchange Format (CLIF) and the Common Logic Standard (CL),
with additions to make working with simple type systems easier.

Logical axioms are called [sentences](#typedlogic.datamodel.Sentence) which organized into [theories](#typedlogic.datamodel.Theory).,
 which can be loaded into a [solver](/integrations/solvers).

 While one of the goals of typed-logic is to be able to write logic intuitively in Python,
 this data model is independent of the mapping from the Python language to the logic language;
 it can be used independently of the python syntax.

Here is an example:

        >>> from typedlogic import Term, Forall, Implies
        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> pdef = PredicateDefinition(predicate='FriendOf',
        ...                          arguments={'x': 'str', 'y': 'str'}),
        >>> theory = Theory(
        ...     name="My theory",
        ...     predicate_definitions=[pdef],
        ... )
        >>> s = Forall([x, y],
        ...            Implies(Term('friend_of', x, y),
        ...                    Term('friend_of', y, x)))
        >>> theory.add(s)

"""
import operator
import types
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type, Union, _SpecialForm, get_origin

SExpressionAtom = Any
SExpressionTerm = List["SExpression"]
SExpression = Union[SExpressionTerm, SExpressionAtom]

@dataclass
class PredicateDefinition:
    """
    Defines the name and arguments of a predicate.

        >>> pdef = PredicateDefinition(predicate='FriendOf',
        ...                            arguments={'x': 'str', 'y': 'str'})

    The arguments are mappings between variable names and types.
    You can use either base types (e.g. 'str', 'int', 'float') or custom types.

    Custom types should be defined in the theory's `type_definitions` attribute.

        >>> pdef = PredicateDefinition(predicate='FriendOf',
        ...                            arguments={'x': 'Person', 'y': 'Person'})
        >>> theory = Theory(
        ...     name="My theory",
        ...     type_definitions={'Person': 'str'},
        ...     predicate_definitions=[pdef],
        ... )

    ```mermaid
    classDiagram
    class PredicateDefinition {
        +String predicate
        +Dict arguments
        +String description
        +Dict metadata
    }
    PredicateDefinition --> "*" PredicateDefinition : parents
    ```

    """

    predicate: str
    arguments: Dict[str, str]
    description: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    parents: Optional[List[str]] = None
    python_class: Optional[Type] = None

    def argument_base_type(self, arg: str) -> str:
        typ = self.arguments[arg]
        try:
            import pydantic
            if isinstance(typ, pydantic.fields.FieldInfo):
                typ = typ.annotation
        except ImportError:
            pass
        return str(typ)

    @classmethod
    def from_class(cls, python_class: Type) -> "PredicateDefinition":
        """
        Create a predicate definition from a python class

        :param predicate_class:
        :return:
        """
        return PredicateDefinition(predicate=python_class.__name__,
                                   arguments={k: v for k, v in python_class.__annotations__.items()},
                                   )


@dataclass
class Variable:
    """
    A variable in a logical sentence.

        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> s = Forall([x, y],
        ...            Implies(Term('friend_of', x, y),
        ...                    Term('friend_of', y, x)))

    Variables can have domains (types) specified:

        >>> x = Variable('x', domain='str')
        >>> y = Variable('y', domain='str')
        >>> z = Variable('y', domain='int')
        >>> xa = Variable('xa', domain='int')
        >>> ya = Variable('ya', domain='int')
        >>> s = Forall([x, y, z],
        ...            Implies(And(Term('ParentOf', x, y),
        ...                        Term('Age', x, xa),
        ...                        Term('Age', y, ya)),
        ...                    Term('OlderThan', x, y)))

        The domains should be either base types or defined types in the theory's `type_definitions` attribute.

    """

    name: str
    domain: Optional[str] = None
    constraints: Optional[List[str]] = None

    def __eq__(self, other):
        return isinstance(other, Variable) and self.name == other.name

    def __str__(self):
        return "?" + self.name

    def __hash__(self):
        return hash(self.name)

    def as_sexpr(self) -> SExpression:
        sexpr = [type(self).__name__, self.name]
        if self.domain:
            return sexpr + [self.domain]
        else:
            return sexpr


class Sentence(ABC):
    """
    Base class for logical sentences.

    Do not use this class directly; use one of the subclasses instead.

    Model:

    ```mermaid
    classDiagram
    Sentence <|-- Term
    Sentence <|-- BooleanSentence
    Sentence <|-- QuantifiedSentence
    Sentence <|-- Extension
    ```


    """

    def __init__(self):
        self._annotations = {}

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def __sub__(self):
        return NegationAsFailure(self)

    def __rshift__(self, other):
        return Implies(self, other)

    def __lshift__(self, other):
        return Implied(self, other)

    def __xor__(self, other):
        return Xor(self, other)

    def iff(self, other):
        return Iff(self, other)

    def __lt__(self, other):
        return Term(operator.lt.__name__, self, other)

    def __le__(self, other):
        return Term(operator.le.__name__, self, other)

    def __gt__(self, other):
        return Term(operator.gt.__name__, self, other)

    def __ge__(self, other):
        return Term(operator.ge.__name__, self, other)

    def __add__(self, other):
        return Term(operator.add.__name__, self, other)

    @property
    def annotations(self):
        return self._annotations

    def as_sexpr(self) -> SExpression:
        raise NotImplementedError(f"type = {type(self)} // {self}")

    @property
    def arguments(self) -> List[Any]:
        raise NotImplementedError(f"type = {type(self)} // {self}")

def as_sexpr(s: Any) -> SExpression:
    if isinstance(s, (Sentence, Variable)):
        return s.as_sexpr()
    if isinstance(s, (Theory, SentenceGroup, PredicateDefinition)):
        sexpr: List[SExpression] = [type(s).__name__]
        for k, v in vars(s).items():
            sexpr.append([k, as_sexpr(v)])
        return sexpr
    if isinstance(s, (list, tuple)):
        return [as_sexpr(x) for x in s]
    if isinstance(s, dict):
        return ["dict", [as_sexpr(x) for x in s.items()]]
    if isinstance(s, Enum):
        return s.value
    if isinstance(s, (type, types.GenericAlias, _SpecialForm)):
        return str(s)
    if hasattr(s, '__origin__') and get_origin(s) is not None:
        return str(s)
    else:
        return s


class Term(Sentence):
    """
    An atomic part of a sentence.

    A ground term is a term with no variables:

        >>> t = Term('FriendOf', 'Alice', 'Bob')
        >>> t
        FriendOf(Alice, Bob)
        >>> t.values
        ('Alice', 'Bob')
        >>> t.is_ground
        True

    Keyword argument based initializaton is also supported:

        >>> t = Term('FriendOf', dict(about='Alice', friend='Bob'))
        >>> t.values
        ('Alice', 'Bob')
        >>> t.positional
        False

    Mappings:

     - Corresponds to AtomicSentence in Common Logic
    """

    def __init__(self, predicate: str, *args, **kwargs):
        self.predicate = predicate
        if not args:
            self.positional = None
            bindings = {}
        elif len(args) == 1 and isinstance(args[0], dict):
            bindings = args[0]
            self.positional = False
        else:
            bindings = {f'arg{i}': arg for i, arg in enumerate(args)}
            self.positional = True
        self.bindings = bindings
        self._annotations = kwargs

    @property
    def is_constant(self):
        """
        :return: True if the term is a constant (zero arguments)
        """
        return not self.bindings

    @property
    def is_ground(self):
        """
        :return: True if none of the arguments are variables
        """
        return not any(isinstance(v, Variable) for v in self.bindings.values())

    @property
    def values(self) -> Tuple[Any, ...]:
        """
        Representation of the arguments of the term as a fixed-position tuples
        :return:
        """
        return tuple([v for v in self.bindings.values()])

    @property
    def variables(self) -> List[Variable]:
        """
        :return: All of the arguments that are variables
        """
        return [v for v in self.bindings.values() if isinstance(v, Variable)]

    @property
    def variable_names(self) -> List[str]:
        return [v.name for v in self.bindings.values() if isinstance(v, Variable)]

    def make_keyword_indexed(self, keywords: List[str]):
        """
        Convert positional arguments to keyword arguments
        """
        if self.positional:
            self.bindings = {k: v for k, v in zip(keywords, self.bindings.values(), strict=False)}
            self.positional = False

    def __repr__(self):
        if not self.bindings:
            return f'{self.predicate}'
        elif self.positional:
            return f'{self.predicate}({", ".join(f"{v}" for v in self.bindings.values())})'
        else:
            return f'{self.predicate}({", ".join(f"{v}" for k, v in self.bindings.items())})'

    def __eq__(self, other):
        # return isinstance(other, Term) and self.predicate == other.predicate and self.bindings == other.bindings
        return isinstance(other, Term) and self.predicate == other.predicate and self.values == other.values

    def __hash__(self):
        return hash((self.predicate, tuple(self.values)))

    def as_sexpr(self) -> SExpression:
        return [self.predicate] + [as_sexpr(v) for v in self.bindings.values()]



class Extension(Sentence, ABC):
    """
    Use this class for framework-specific extensions.

    An example of this is the `Fact` class which subclasses Extension, and is intended to be
    subclasses by domain-specific classes representing predicate definitions, whose instances
    map to terms.
    """

    @abstractmethod
    def to_model_object(self) -> Sentence:
        """
        Convert the extension to a standard model object.

        :return:
        """
        pass

    def as_sexpr(self) -> List[Any]:
        return self.to_model_object().as_sexpr()

    @property
    def arguments(self) -> List[Any]:
        return self.to_model_object().arguments

def term(predicate: Union[str, Type[Extension], Extension], *args, **kwargs) -> Term:
    """
    Create a term object.

    :param predicate:
    :param args:
    :param kwargs:
    :return: Term object
    """
    if isinstance(predicate, Extension):
        s = predicate.to_model_object()
        if isinstance(s, Term):
            return s
        else:
            raise ValueError(f"Cannot convert {predicate} to a Term")
    if isinstance(predicate, type):
        predicate = predicate.__name__
    return Term(predicate, *args, **kwargs)

@dataclass
class BooleanSentence(Sentence, ABC):
    """
    Base class for sentences that are boolean expressions

    Corresponds to BooleanSentence in CL

    ```mermaid
    classDiagram
    BooleanSentence <|-- And
    BooleanSentence <|-- Or
    BooleanSentence <|-- Not
    BooleanSentence <|-- Xor
    BooleanSentence <|-- ExactlyOne
    BooleanSentence <|-- Implication
    BooleanSentence <|-- Implied
    BooleanSentence <|-- Iff
    BooleanSentence <|-- NegationAsFailure
    BooleanSentence --> "*" Sentence : operands
    ```

    """

    operands: Tuple = field(default_factory=tuple)

    def __init__(self, *operands, **kwargs):
        self.operands = operands
        self._annotations = kwargs

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.operands == other.operands

    def as_sexpr(self) -> SExpression:
        return [type(self).__name__] + [as_sexpr(op) for op in self.operands]

    @property
    def arguments(self) -> List[Any]:
        return list(self.operands)


@dataclass
class And(BooleanSentence):
    """
    A conjunction of sentences.

        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> s = And(Term('friend_of', x, y), Term('friend_of', y, x))

    You can also use syntactic sugar:

        >>> s = Term('friend_of', x, y) & Term('friend_of', y, x)

    Note however that precedence rules for `&` are different from `and`.

    In the context of a pylog program, you can also use `and`:

    ```assert FriendOf(x, y) & FriendOf(y, x)```

    As in CL, ``And()`` means True
    """

    def __init__(self, *operands, **kwargs):
        super().__init__(*operands, **kwargs)


    def __str__(self):
        return f'({") & (".join(str(op) for op in self.operands)})'

    def __repr__(self):
        return f'And({", ".join(repr(op) for op in self.operands)})'

@dataclass
class Or(BooleanSentence):
    """
    A disjunction of sentences.

        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> s = Or(Term('friend_of', x, y), Term('friend_of', y, x))
        >>> s.operands[0]
        friend_of(?x, ?y)

    You can also use syntactic sugar:

        >>> s = Term('friend_of', x, y) | Term('friend_of', y, x)

    Note however that precedence rules for `|` are different from `or`.

    In the context of a pylog program, you can also use `or`:

    ```assert FriendOf(x, y) | FriendOf(y, x)```

    As in CL, ``Or()`` means False
    """

    def __init__(self, *operands, **kwargs):
        super().__init__(*operands, **kwargs)

    def __str__(self):
        return f'({") | (".join(str(op) for op in self.operands)})'

    def __repr__(self):
        return f'Or({", ".join(repr(op) for op in self.operands)})'

@dataclass
class Not(BooleanSentence):
    """
    A complement of a sentence

        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> s = Not(Term('friend_of', x, y))
        >>> s.negated
        friend_of(?x, ?y)

    You can also use syntactic sugar:

        >>> s = ~Term('friend_of', x, y)

    In the context of a pylog program, you can also use `not`:

    ```assert not FriendOf(x, y)```

    This SHOULD be interpreted as strict negation, not as failure.
    """

    def __init__(self, operand, **kwargs):
        super().__init__(operand, **kwargs)

    def __str__(self):
        return f'~{self.operands[0]}'

    def __repr__(self):
        return f'Not({repr(self.operands[0])})'

    @property
    def negated(self) -> Sentence:
        """
        The negated sentence
        :return: Sentence
        """
        return self.operands[0]

class Xor(BooleanSentence):
    """
    An exclusive or of sentences
    """

    def __init__(self, left, right, **kwargs):
        super().__init__(left, right, **kwargs)


@dataclass
class ExactlyOne(BooleanSentence):
    """
    Exactly one of the sentences is true

        >>> x = Variable('x')
        >>> s = ExactlyOne(Term('likes', x, "root beer"), Term('likes', x, "marmite"))

    """

    def __init__(self, *operands, **kwargs):
        super().__init__(*operands, **kwargs)

    def __str__(self):
        return f'({") x| (".join(str(op) for op in self.operands)})'

    def __repr__(self):
        return f'ExactlyOne({", ".join(repr(op) for op in self.operands)})'


@dataclass
class Implication(BooleanSentence, ABC):
    """
    An abstract grouping of sentences with an implication operator.

    ```mermaid
    classDiagram
    Implication <|-- Implies
    Implication <|-- Implied
    Implication <|-- Iff
    ```

    """

    def __init__(self, left, right, **kwargs):
        super().__init__(left, right, **kwargs)

    @property
    def symbol(self):
        raise NotImplementedError

    def __str__(self):
        return f'({self.operands[0]} {self.symbol} {self.operands[1]})'

    def __repr__(self):
        return f'{type(self).__name__}({repr(self.operands[0])}, {repr(self.operands[1])})'

@dataclass
class Implies(Implication):
    """
    An if-then implication.

    Corresponds to Implication in CommonLogic

        >>> x = Variable('x')
        >>> s = Iff(Term('likes', x, "root beer"), ~Term('likes', x, "marmite"))

    You can also use syntactic sugar:

        >>> s = Term('likes', x, "root beer") >> ~Term('likes', x, "marmite")

    """

    def __init__(self, antecedent, consequent, **kwargs):
        super().__init__(antecedent, consequent, **kwargs)

    @property
    def symbol(self):
        return "->"

    @property
    def antecedent(self):
        return self.operands[0]

    @property
    def consequent(self):
        return self.operands[1]

    def __str__(self):
        return f'({self.operands[0]} -> {self.operands[1]})'

    def __repr__(self):
        return f'Implies({repr(self.operands[0])}, {repr(self.operands[1])})'


@dataclass
class Implied(Implication):
    """
    An implication of the form consequent <- antecedent

    Inverse of `Implies`
    """

    def __init__(self, consequent, antecedent, **kwargs):
        super().__init__(consequent, antecedent, **kwargs)

    @property
    def symbol(self):
        return "<-"

    @property
    def antecedent(self):
        return self.operands[1]

    @property
    def consequent(self):
        return self.operands[0]


    def __str__(self):
        return f'({self.operands[0]} <- {self.operands[1]})'

    def __repr__(self):
        return f'Implied({repr(self.operands[0])}, {repr(self.operands[1])})'


@dataclass
class Iff(Implication):
    """
    An equivalence of sentences

    Corresponds to Biconditional in CommonLogic.

        >>> x = Variable('x')
        >>> s = Iff(Term('likes', x, "jaffa cakes"), Term('likes', x, "marmite"))

    """

    def __init__(self, left, right, **kwargs):
        super().__init__(left, right, **kwargs)

    @property
    def symbol(self):
        return "<->"

    @property
    def left(self):
        return self.operands[0]

    @property
    def right(self):
        return self.operands[1]

    def __str__(self):
        return f'({self.operands[0]} <-> {self.operands[1]})'

    def __repr__(self):
        return f'Iff({repr(self.operands[0])}, {repr(self.operands[1])})'


@dataclass
class NegationAsFailure(BooleanSentence):
    """
    A negated sentence, interpreted via negation as failure semantics.
    """

    def __init__(self, operand, **kwargs):
        super().__init__(operand, **kwargs)

    @property
    def negated(self):
        return self.operands[0]

# deprecate this?
def not_provable(predicate):
    """Function for Negation as Failure"""
    return NegationAsFailure(predicate)


@dataclass
class QuantifiedSentence(Sentence, ABC):
    """
    A sentence with a logical quantifier.

    ```mermaid
    classDiagram
    QuantifiedSentence <|-- Forall
    QuantifiedSentence <|-- Exists
    QuantifiedSentence --> "*" Variable : variables
    ```

    """

    variables: List[Variable]
    sentence: Sentence
    _annotations: Optional[Dict[str, Any]] = None

    @property
    def quantifier(self) -> str:
        raise NotImplementedError

    def _bindings_str(self) -> str:
        return ", ".join(f"{v.name}: {v.domain}" for v in self.variables)

    def __hash__(self):
        return hash((self.quantifier, tuple(self.variables), self.sentence))

    def as_sexpr(self) -> SExpression:
        return [type(self).__name__, [v.as_sexpr() for v in self.variables], self.sentence.as_sexpr()]

    @property
    def arguments(self) -> List[Any]:
        return [self.variables, self.sentence]



@dataclass
class Forall(QuantifiedSentence):
    """
    Universal quantifier.

        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> s = Forall([x, y], Implies(Term('friend_of', x, y), Term('friend_of', y, x)))

    """

    @property
    def quantifier(self) -> str:
        return "all"

    def __str__(self):
        return f'∀{self._bindings_str()} : {self.sentence}'

    def __repr__(self):
        return f'Forall([{self._bindings_str()}] : {repr(self.sentence)})'


@dataclass
class Exists(QuantifiedSentence):
    """
    Existential quantifier.

        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> s = ~Exists([x, y], And(Term('friend_of', x, y), Term('enemy_of', x, y)))

    """

    @property
    def quantifier(self) -> str:
        return "exists"

    def __str__(self):
        return f'∃{self.sentence}'

    def __repr__(self):
        return f'Exists({self._bindings_str()} : {repr(self.sentence)})'

    def __hash__(self):
        return hash((self.quantifier, self._bindings_str(), self.sentence))


class SentenceGroupType(str, Enum):
    AXIOM = "axiom"
    GOAL = "goal"

@dataclass
class SentenceGroup:
    """
    A logical grouping of related sentences with common documentation.

    One way to collect these is via a decorated python function.

    ```mermaid
    classDiagram
    class SentenceGroup {
        +String name
        +SentenceGroupType group_type
        +String docstring
        +Dict annotations
    }
    SentenceGroup --> "*" Sentence : sentences

    """

    name: str
    group_type: Optional[SentenceGroupType] = None
    docstring: Optional[str] = None
    sentences: Optional[List[Sentence]] = None
    _annotations: Optional[Dict[str, Any]] = None

DefinedType = Union["DefinedUnionType", str]
DefinedUnionType = List[DefinedType]

@dataclass
class Theory:
    """
    A collection of predicate definitions and sentences.

    Analogous to a Text in CommonLogic.

    ```mermaid
    classDiagram
    class Theory {
        +String name
        +Dict constants
        +Dict type_definitions
        +List predicate_definitions
        +List sentence_groups
        +List ground_terms
        +Dict annotations
    }
    Theory --> "*" DefinedType : type_definitions
    Theory --> "*" PredicateDefinition : predicate_definitions
    Theory --> "*" Term : ground_terms
    Theory --> "*" SentenceGroup : sentence_groups
    ```

    """

    name: Optional[str] = None
    constants: Dict[str, Any] = field(default_factory=dict)
    type_definitions: Dict[str, DefinedType] = field(default_factory=dict)
    predicate_definitions: List[PredicateDefinition] = field(default_factory=list)
    sentence_groups: List[SentenceGroup] = field(default_factory=list)
    ground_terms: List[Term] = field(default_factory=list)
    _annotations: Optional[Dict[str, Any]] = None
    source_module_name: Optional[str] = None

    @property
    def sentences(self) -> List[Sentence]:
        """
        Return all sentences in the theory

        :return:
        """
        if self.sentence_groups:
            return [s for sg in self.sentence_groups for s in sg.sentences or []]
        return []

    @property
    def goals(self) -> List[Sentence]:
        """
        Return all goal sentences in the theory

        :return:
        """
        return [s for sg in self.sentence_groups or [] if sg.group_type == SentenceGroupType.GOAL for s in sg.sentences or []]

    def add(self, sentence: Sentence):
        """
        Add a sentence to the theory

        :param sentence:
        :return:
        """
        if isinstance(sentence, Extension):
            sentence = sentence.to_model_object()
        if not self.sentence_groups:
            self.sentence_groups = []
        self.sentence_groups.append(SentenceGroup(name="Sentences", sentences=[sentence]))

    def remove(self, sentence: Sentence, strict=False):
        """
        Remove a sentence to the theory

        :param sentence:
        :param strict:
        :return:
        """
        if isinstance(sentence, Extension):
            sentence = sentence.to_model_object()
        if not self.sentence_groups:
            if strict:
                raise ValueError("No sentences to remove from")
            return
        n = 0
        for sg in self.sentence_groups:
            if sg.sentences and sentence in sg.sentences:
                sg.sentences.remove(sentence)
                n += 1
        if n != 1 and strict:
            raise ValueError(f"Removed {n} sentences")

    def unroll_type(self, typ: DefinedType) -> List[str]:
        """
        Unroll a defined type into its components

        :param typ:
        :return:
        """
        if isinstance(typ, str):
            if typ in self.type_definitions:
                return self.unroll_type(self.type_definitions[typ])
            return [typ]
        if isinstance(typ, list):
            ts = []
            for t in typ:
                ts.extend(self.unroll_type(t))
            return ts
        raise ValueError(f"Unknown type {typ}")


def as_object(self: Any, parent=None) -> Any:
    if parent == "constants":
        # TODO
        return {}
    if isinstance(self, Extension):
        return as_object(self.to_model_object())
    elif isinstance(self, Term):
        return {
            "type": type(self).__name__,
            "arguments": [self.predicate] + [as_object(x) for x in self.values]
        }
    elif isinstance(self, Variable):
        return {
            "type": type(self).__name__,
            "arguments": self.as_sexpr()[1:]
        }
    elif isinstance(self, Sentence):
        return {
            "type": type(self).__name__,
            "arguments": [as_object(x) for x in self.arguments]
        }
    elif isinstance(self, (PredicateDefinition, Theory, SentenceGroup)):
        return {
            "type": type(self).__name__,
            **{k: as_object(v, k) for k, v in vars(self).items() if v is not None},
        }
    elif isinstance(self, Enum):
        return self.value
    elif isinstance(self, list):
        return [as_object(x) for x in self]
    else:
        return self

def from_object(obj: Any) -> Any:
    if isinstance(obj, dict):
        if "type" in obj:
            cls = globals()[obj["type"]]
            if cls in (PredicateDefinition, Theory, SentenceGroup):
                return cls(**{k: from_object(v) for k, v in obj.items() if k != "type"})
            else:
                return cls(*[from_object(x) for x in obj["arguments"]])
        else:
            return obj
    if isinstance(obj, list):
        return [from_object(x) for x in obj]
    return obj

class NotInProfileError(ValueError):
    """
    Raised when a sentence is not in some profile
    """

    pass



"""
Function for performing transformation and manipulation of Sentences and Theories
"""
import json
from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field
from typing import List, Iterator, Iterable, Mapping, Callable, Optional, Dict, Any, Union, Type, Tuple

from typedlogic import Theory, Forall, Variable, Implies, Term, SentenceGroup, Sentence, BooleanSentence, \
    NegationAsFailure
from typedlogic.builtins import NUMERIC_BUILTINS, NAME_TO_INFIX_OP
from typedlogic.datamodel import Iff, And, Implied, Or, QuantifiedSentence, Not, Exists, Extension, Xor, ExactlyOne, \
    NotInProfileError
from typedlogic.utils.detect_stratified_negation import analyze_datalog_program


def sentences_from_predicate_hierarchy(theory: Theory) -> List[Sentence]:
    new_sentences: List[Sentence] = []
    if not theory.predicate_definitions:
        raise ValueError("Theory must have predicate definitions")
    for pd in theory.predicate_definitions:
        if pd.parents:
            for parent in pd.parents:
                # bindings = {arg: Variable(arg, domain=typ) for arg, typ in pd.arguments.items()}
                vars = [Variable(arg, domain=typ) for arg, typ in pd.arguments.items()]
                args = [Variable(arg) for arg in pd.arguments]
                impl = Implies(
                        antecedent=Term(parent, *args),
                        consequent=Term(pd.predicate, *args),
                    )
                qs = Forall(vars, impl)
                if qs not in theory.sentences:
                    new_sentences.append(qs)

    return new_sentences

def implies_from_parents(theory: Theory) -> Theory:
    """
    Generate implications from parent classes.

    In the data model, PredicateDefinition can be linked to any number
    of parent classes. This function generates implications for each.

        >>> from typedlogic import PredicateDefinition, Theory, SentenceGroup
        >>> theory = Theory(
        ...     name="test",
        ...     predicate_definitions=[
        ...         PredicateDefinition(
        ...             predicate="Person",
        ...             arguments={"name": "str"},
        ...             parents=["Thing"],
        ...         ),
        ...         PredicateDefinition(
        ...             predicate="Thing",
        ...             arguments={"name": "str"},
        ...             parents=[],
        ...         ),
        ...     ],
        ...     sentence_groups=[]
        ... )
        >>> theory2 = implies_from_parents(theory)
        >>> new_sentences = theory2.sentence_groups[0].sentences
        >>> new_sentences
        [Forall([name: str] : Implies(Thing(?name), Person(?name)))]

    :param theory: The theory to generate implications for
    :return: A new theory with implications
    """
    new_sentences = sentences_from_predicate_hierarchy(theory)
    new_sg = SentenceGroup(name="Inferred", sentences=new_sentences)
    sgs = (theory.sentence_groups or []) + [new_sg]
    return Theory(
        name=theory.name,
        predicate_definitions=theory.predicate_definitions,
        sentence_groups=sgs,
    )

def disjunction_as_list(sentence: Sentence) ->List[Sentence]:
    if isinstance(sentence, Or):
        return list(sentence.operands)
    return [sentence]

def conjunction_as_list(sentence: Sentence) -> List[Sentence]:
    if isinstance(sentence, And):
        return list(sentence.operands)
    return [sentence]

@dataclass
class PrologConfig:
    """
    Configuration for Prolog output.
    """
    use_lowercase_vars: Optional[bool] = False
    use_uppercase_predicates: Optional[bool]  = False
    disjunctive_datalog: Optional[bool]  = False
    operator_map: Optional[Mapping[str, str]] = None
    negation_symbol : str = field(default=r'\+')
    negation_as_failure_symbol : str = field(default=r'\+')
    assume_negation_as_failure: bool = False
    double_quote_strings: bool = False
    include_parens_for_zero_args: bool = False
    allow_function_terms: bool = True
    allow_nesting: bool = True
    null_term: str = "null(_)"
    allow_skolem_terms: bool = False

def as_prolog(sentence: Union[Sentence, List[Sentence]], config: Optional[PrologConfig]=None, depth=0, translate=False, strict=True) -> str:
    """
    Convert a sentence to Prolog syntax.

        >>> from typedlogic import Term, Variable, Forall
        >>> X = Variable("x", "str")
        >>> Y = Variable("y", "str")
        >>> A = Term("A", X, Y)
        >>> B = Term("B", X, Y, "foo", 5)
        >>> as_prolog(A)
        'a(X, Y)'

        >>> as_prolog(Forall([X], A))
        'a(X, Y)'

        >>> C = Term("C")
        >>> D = Term("D")
        >>> E = Term("E")
        >>> print(as_prolog(Implies(C, D)))
        d :- c.

        >>> print(as_prolog(Implies(C & E, D)))
        d :- c, e.

        >>> print(as_prolog(Implies(And(D, E), C)))
        c :- d, e.

        >>> print(as_prolog(Implies(Or(D, E), C)))
        c :- (d; e).

        >>> config = PrologConfig(disjunctive_datalog=True)
        >>> print(as_prolog(Implies(And(E), C | D), config, depth=1))
        c; d :- e.

        >>> config.negation_symbol = '!'
        >>> as_prolog(~A, config, depth=1)
        '! (a(X, Y))'

        >>> print(as_prolog(Term('lt', X, 5)))
        X < 5

        >>> print(as_prolog(Term('eq', X, 5)))
        X == 5

        >>> print(as_prolog(Implies(And(C, Exists([X], Term("A", X))), D)))
        d :- c, a(X).

    :param sentence: the sentence to render
    :param config:
    :param depth:
    :param translate:
    :param strict:
    :return:
    """
    if isinstance(sentence, list):
        return "\n".join(as_prolog(s, config, depth=depth) for s in sentence)
    if not config:
        config = PrologConfig()
    def _paren(s: str) -> str:
        if config.allow_nesting:
            return f"({s})"
        return s
    if translate:
        rules = to_horn_rules(sentence, allow_disjunctions_in_head=config.disjunctive_datalog)
        return "\n".join(as_prolog(s, config, depth=depth) for s in rules)
    if isinstance(sentence, Forall):
        sentence = sentence.sentence
    if depth == 0 and not isinstance(sentence, (Implies, Term)):
        raise NotInProfileError(f"Top level sentence must be an implication or term {sentence}, got {type(sentence)}")
    if isinstance(sentence, Exists) and depth > 0:
        sentence = sentence.sentence
    if isinstance(sentence, And):
        if not sentence.operands:
            return "true"
        return f"{', '.join(as_prolog(op, config, depth+1) for op in sentence.operands)}"
    if isinstance(sentence, Or):
        if not sentence.operands:
            return "fail"
        return _paren(f"{'; '.join(as_prolog(op, config, depth+1) for op in sentence.operands)}")
    if isinstance(sentence, (Not, NegationAsFailure)):
        negated_clause = _paren(as_prolog(sentence.negated, config, depth+1))
        return f"{config.negation_symbol} {negated_clause}"
    if isinstance(sentence, Term):
        if not config.allow_skolem_terms:
            for t in sentence.values:
                if isinstance(t, Term) and t.predicate.startswith("sk__"):
                    raise NotInProfileError(f"Skolem term not supported: {sentence}")
        vals = list(sentence.bindings.values())
        def _render_arg(v: Any) -> str:
            if v is None:
                if depth > 0:
                    return "_"
                else:
                    return config.null_term
            if isinstance(v, Variable):
                if config.use_lowercase_vars:
                    return v.name
                return v.name.capitalize()
            if isinstance(v, Term):
                if not config.allow_function_terms:
                    raise ValueError(f"Nested term not supported: {v}")
                return as_prolog(v, config, depth+1)
            if config.double_quote_strings:
                return json.dumps(v)
            else:
                return repr(v)
        p = sentence.predicate
        operator_map = {k: v for k, v in NAME_TO_INFIX_OP.items()}
        if config.operator_map:
            operator_map.update(config.operator_map)
        if p in operator_map:
            p = operator_map[p]
            if len(vals) == 2:
                return f"{_render_arg(vals[0])} {p} {_render_arg(vals[1])}"
            elif len(vals) == 1:
                return f"{p} {_render_arg(vals[0])}"
            else:
                raise ValueError(f"Operator {p} only supports 1 or 2 arguments")
        else:
            if config.use_uppercase_predicates:
                p = p.capitalize()
            if config.use_uppercase_predicates is False:
                p = p.lower()
            if not vals and not config.include_parens_for_zero_args:
                return p
            else:
                return f"{p}({', '.join([_render_arg(v) for v in vals])})"
    if not isinstance(sentence, Implies):
        raise NotInProfileError(f"Unsupported sentence {sentence}")
    # assumption: generation a (head :- body) implication
    if isinstance(sentence.consequent, Or) and len(sentence.consequent.operands) > 1:
        if not config.disjunctive_datalog:
            raise NotInProfileError(f"Disjunctions on LHS not allowed {sentence}")
    if isinstance(sentence.consequent, And):
        raise NotInProfileError(f"Conjunctions on LHS not allowed {sentence}\n"
                                "Transform using simplify_prolog_transform first")
    # check for unbound variables
    body_vars = []
    # eliminate Exists
    antecedent_list = [t.sentence if isinstance(t, Exists) else t for t in conjunction_as_list(sentence.antecedent)]
    for body_term in antecedent_list:
        if not isinstance(body_term, Term):
            # TODO: this currently assumes disjunctions are unrolled from body
            # raise NotInProfileError(f"Body must be a term {sentence}")
            continue
        body_vars.extend(body_term.variable_names)
    for head_term in disjunction_as_list(sentence.consequent):
        if isinstance(head_term, Not):
            continue
        if not isinstance(head_term, Term):
            raise NotInProfileError(f"Head must be a term, got: {type(head_term)} in {sentence}")
        head_vars = head_term.variable_names
        for v in head_vars:
            if v not in body_vars:
                raise NotInProfileError(f"Variable {v} in head not in body {sentence}")

    head = as_prolog(sentence.consequent, config, depth+1)
    body = as_prolog(sentence.antecedent, config, depth+1)
    if head.startswith("(") and head.endswith(")"):
        head = head[1:-1]
    if body == "true":
        return f"{head}."
    if head == "fail":
        return f":- {body}."
    return f"{head} :- {body}."


def simple_prolog_transform(sentence: Sentence, strict=False) -> List[Sentence]:
    """
    Transform a sentence to a list of sentences suitable for Prolog.

    The resulting sentences will all be quantified Body -> Head implications,
    where Body is a conjunction

        >>> from typedlogic import Term, Variable, Forall
        >>> A = Term("A")
        >>> B = Term("B")
        >>> C = Term("C")
        >>> simple_prolog_transform(A >> B)
        [Forall([] : Implies(And(A), B))]

        >>> simple_prolog_transform(Iff(A, B))
        [Forall([] : Implies(And(B), A)), Forall([] : Implies(And(A), B))]

        >>> for s in simple_prolog_transform(Iff(A, B)):
        ...     print(as_prolog(s))
        a :- b.
        b :- a.

        >>> for s in simple_prolog_transform((And(A) & And(B)) >> C):
        ...     print(as_prolog(s))
        c :- a, b.

        >>> for s in simple_prolog_transform((Or(A,B)) >> C):
        ...     print(as_prolog(s))
        c :- b.
        c :- a.

        >>> for s in simple_prolog_transform(C >> (A & B)):
        ...     print(as_prolog(s))
        b :- c.
        a :- c.




    :param sentence:
    :param strict:
    :return:
    """
    def not_in_profile(s: Sentence) -> None:
        if strict:
            raise NotInProfileError(f"Unsupported sentence {s}")
    sentence = transform_sentence(sentence, reduce_singleton)
    sentence = transform_sentence(sentence, eliminate_iff)
    if not isinstance(sentence, Forall):
        sentence = Forall([], sentence)
    # sentence = transform_sentence(sentence, eliminate_implies)
    outer = sentence
    sentence = sentence.sentence
    if isinstance(sentence, And):
        # expand And to multiple sentences
        sentences = [op for op in sentence.operands]
    else:
        sentences = [sentence]
    new_sentences: List[Sentence] = []
    while sentences:
        sentence = sentences.pop()
        if isinstance(sentence, Term):
            new_sentences.append(sentence)
            continue
        if isinstance(sentence, Implied):
            sentences.append(Implies(sentence.operands[1], sentence.operands[0]))
        if isinstance(sentence, Iff):
            sentences.append(Implies(sentence.left, sentence.right))
            sentences.append(Implies(sentence.right, sentence.left))
        if not isinstance(sentence, Implies):
            not_in_profile(sentence)
            continue
        body = sentence.antecedent
        head = sentence.consequent
        if isinstance(head, And):
            sentences.extend([Implies(body, op) for op in head.operands])
            continue
        if not isinstance(head, Term):
            not_in_profile(sentence)
            continue
        if isinstance(body, Or):
            sentences.extend([Implies(op, head) for op in body.operands])
            continue
        if isinstance(body, Term):
            body = And(body)
        if not isinstance(body, And):
            # note: does not check members of And
            not_in_profile(sentence)
            continue
        new_sentences.append(Implies(body, head))
    return [Forall(outer.variables, s) for s in new_sentences]


def as_fol(sentence, config: Optional[PrologConfig]=None) -> str:
    """
    Convert a sentence to first order logic syntax.

    >>> from typedlogic import Term, Variable, Forall
    >>> X = Variable("x", "str")
    >>> Y = Variable("y", "str")
    >>> A = Term("A", X)
    >>> B = Term("B", X)
    >>> print(as_fol(Forall([X], A >> B)))
    ∀[x:str]. A(x) → B(x)
    >>> print(as_fol(Forall([X], Iff(A, B))))
    ∀[x:str]. A(x) ↔ B(x)
    >>> print(as_fol(Exists([X], A & B)))
    ∃[x:str]. A(x) ∧ B(x)
    >>> print(as_fol(A|B))
    (A(x) ∨ B(x))

    :param sentence:
    :param config:
    :return:
    """
    if not config:
        config = PrologConfig(use_lowercase_vars=True, use_uppercase_predicates=None)
    if isinstance(sentence, (Exists, Forall)):
        qsym = "∀" if isinstance(sentence, Forall) else "∃"
        arg_exprs = [f"{v.name}:{v.domain}" if v.domain else v.name for v in sentence.variables]
        args = " ".join(arg_exprs)
        return f"{qsym}[{args}]. {as_fol(sentence.sentence, config)}"
    if isinstance(sentence, And):
        return f"{' ∧ '.join(as_fol(op, config) for op in sentence.operands)}"
    if isinstance(sentence, Or):
        return f"({' ∨ '.join(as_fol(op, config) for op in sentence.operands)})"
    if isinstance(sentence, Not):
        return f"¬{as_fol(sentence.negated, config)}"
    if isinstance(sentence, Implies):
        return f"{as_fol(sentence.antecedent, config)} → {as_fol(sentence.consequent, config)}"
    if isinstance(sentence, Implied):
        return f"{as_fol(sentence.operands[1], config)} ← {as_fol(sentence.operands[0], config)}"
    if isinstance(sentence, Iff):
        return f"{as_fol(sentence.left, config)} ↔ {as_fol(sentence.right, config)}"
    if isinstance(sentence, Term):
        vals = list(sentence.bindings.values())
        def _render_arg(v: Any) -> str:
            if isinstance(v, Variable):
                if config.use_lowercase_vars:
                    return v.name
                return v.name.capitalize()
            if config.double_quote_strings:
                return json.dumps(v)
            else:
                return repr(v)
        p = sentence.predicate
        operator_map = {k: v for k, v in NAME_TO_INFIX_OP.items()}
        if config.operator_map:
            operator_map.update(config.operator_map)
        if p in operator_map:
            p = operator_map[p]
            if len(vals) == 2:
                return f"{_render_arg(vals[0])} {p} {_render_arg(vals[1])}"
            elif len(vals) == 1:
                return f"{p} {_render_arg(vals[0])}"
            else:
                raise ValueError(f"Operator {p} only supports 1 or 2 arguments")
        else:
            if config.use_uppercase_predicates:
                p = p.capitalize()
            if config.use_uppercase_predicates is False:
                p = p.lower()
            if not vals and not config.include_parens_for_zero_args:
                return p
            else:
                return f"{p}({', '.join([_render_arg(v) for v in vals])})"
    return ""


def as_tptp(sentence: Sentence, config: Optional[PrologConfig] = None, depth=0) -> str:
    """
    Convert a sentence to TPTP syntax.

    :param sentence: The sentence to convert
    :param config: Configuration options (optional)
    :param depth: Current depth in the sentence structure (used for indentation)
    :return: TPTP representation of the sentence

    Examples:
    >>> from typedlogic import Term, Variable, Forall, Exists, And, Or, Not, Implies
    >>> X = Variable("X", "str")
    >>> Y = Variable("Y", "str")
    >>> P = Term("P", X)
    >>> Q = Term("Q", X, Y)
    >>> R = Term("R", Y)
    >>> print(as_tptp(Forall([X], Implies(P, Q))))
    ! [X] : (p(X) => q(X, Y))
    >>> print(as_tptp(Exists([X, Y], And(P, Q))))
    ? [X, Y] : (p(X) & q(X, Y))
    >>> print(as_tptp(Or(P, Not(R))))
    (p(X) | ~r(Y))
    """
    if not config:
        config = PrologConfig(use_lowercase_vars=False, use_uppercase_predicates=False)

    def format_var(v: Variable) -> str:
        return v.name if config.use_lowercase_vars else v.name.capitalize()

    def format_predicate(p: str) -> str:
        return p.lower() if config.use_uppercase_predicates is False else p

    if isinstance(sentence, (Forall, Exists)):
        quantifier = "!" if isinstance(sentence, Forall) else "?"
        vars = ", ".join(format_var(v) for v in sentence.variables)
        return f"{quantifier} [{vars}] : {as_tptp(sentence.sentence, config, depth)}"

    elif isinstance(sentence, And):
        return f"({' & '.join(as_tptp(op, config, depth + 1) for op in sentence.operands)})"

    elif isinstance(sentence, Or):
        return f"({' | '.join(as_tptp(op, config, depth + 1) for op in sentence.operands)})"

    elif isinstance(sentence, Not):
        return f"~{as_tptp(sentence.negated, config, depth + 1)}"

    elif isinstance(sentence, Implies):
        return f"({as_tptp(sentence.antecedent, config, depth + 1)} => {as_tptp(sentence.consequent, config, depth + 1)})"

    elif isinstance(sentence, Iff):
        return f"({as_tptp(sentence.left, config, depth + 1)} <=> {as_tptp(sentence.right, config, depth + 1)})"

    elif isinstance(sentence, Term):
        predicate = format_predicate(sentence.predicate)
        args = ", ".join(format_var(v) if isinstance(v, Variable) else repr(v) for v in sentence.bindings.values())
        return f"{predicate}({args})"

    else:
        raise ValueError(f"Unsupported sentence type: {type(sentence)}")


def tptp_problem(theory: Theory, conjecture: Optional[Sentence] = None) -> str:
    """
    Generate a complete TPTP problem from a theory and an optional conjecture.

    :param theory: The theory containing axioms
    :param conjecture: An optional conjecture to prove
    :return: TPTP representation of the problem

    Example:
    >>> from typedlogic import Theory, PredicateDefinition, Term, Variable, Forall, Implies
    >>> X = Variable("X", "str")
    >>> Y = Variable("Y", "str")
    >>> theory = Theory(
    ...     name="example",
    ...     predicate_definitions=[
    ...         PredicateDefinition("P", {"x": "str"}),
    ...         PredicateDefinition("Q", {"x": "str", "y": "str"}),
    ...     ],
    ...     sentence_groups=[
    ...         SentenceGroup("axioms", sentences=[
    ...             Forall([X], Implies(Term("P", X), Term("Q", X, Y)))
    ...         ])
    ...     ]
    ... )
    >>> conjecture = Forall([X, Y], Implies(Term("P", X), Term("Q", X, Y)))
    >>> print(tptp_problem(theory, conjecture))
    % Problem: example
    fof(axiom1, axiom, ! [X] : (p(X) => q(X, Y))).
    fof(conjecture, conjecture, ! [X, Y] : (p(X) => q(X, Y))).
    """
    lines = [f"% Problem: {theory.name}"]

    for i, sentence in enumerate(theory.sentences, 1):
        lines.append(f"fof(axiom{i}, axiom, {as_tptp(sentence)}).")

    if conjecture:
        lines.append(f"fof(conjecture, conjecture, {as_tptp(conjecture)}).")

    return "\n".join(lines)


def as_prover9(sentence: Sentence, config: Optional[PrologConfig] = None, depth=0) -> str:
    """
    Convert a sentence to Prover9 syntax.

    :param sentence: The sentence to convert
    :param config: Configuration options (optional)
    :param depth: Current depth in the sentence structure (used for indentation)
    :return: Prover9 representation of the sentence

    Examples:
    >>> from typedlogic import Term, Variable, Forall, Exists, And, Or, Not, Implies
    >>> X = Variable("X", "str")
    >>> Y = Variable("Y", "str")
    >>> P = Term("P", X)
    >>> Q = Term("Q", X, Y)
    >>> R = Term("R", Y)
    >>> print(as_prover9(Forall([X], Implies(P, Q))))
    all x ((P(x) -> Q(x, y)))
    >>> print(as_prover9(Exists([X, Y], And(P, Q))))
    exists x y ((P(x) & Q(x, y)))
    >>> print(as_prover9(Or(P, Not(R))))
    (P(x) | - ( R(y) ))
    >>> print(as_prover9(Term("S", "hello")))
    S(s_hello)
    """
    if not config:
        config = PrologConfig(use_lowercase_vars=True, use_uppercase_predicates=False)

    def format_var(v: Variable) -> str:
        return v.name.lower()

    def format_predicate(p: str) -> str:
        return p.upper() if config.use_uppercase_predicates else p

    def format_value(v: Any) -> str:
        if isinstance(v, str):
            # Convert string to a valid Prover9 constant
            return f"s_{v.replace(' ', '_')}"
        elif isinstance(v, int):
            return str(v)
        elif isinstance(v, float):
            # Prover9 doesn't support floats directly, so we convert to a fraction
            from fractions import Fraction
            frac = Fraction(v).limit_denominator()
            return f"rational({frac.numerator},{frac.denominator})"
        elif v is None:
            return "null"
        else:
            return str(v)

    if isinstance(sentence, (Forall, Exists)):
        quantifier = "all" if isinstance(sentence, Forall) else "exists"
        vars = " ".join(format_var(v) for v in sentence.variables)
        return f"{quantifier} {vars} ({as_prover9(sentence.sentence, config, depth)})"

    elif isinstance(sentence, And):
        return f"({' & '.join(as_prover9(op, config, depth + 1) for op in sentence.operands)})"

    elif isinstance(sentence, Xor):
        return as_prover9(expand_xor(sentence), config, depth)

    elif isinstance(sentence, ExactlyOne):
        return as_prover9(expand_exactly_one(sentence), config, depth)

    elif isinstance(sentence, Or):
        return f"({' | '.join(as_prover9(op, config, depth + 1) for op in sentence.operands)})"

    elif isinstance(sentence, Not):
        return f"- ( {as_prover9(sentence.negated, config, depth + 1)} )"

    elif isinstance(sentence, Implies):
        return f"({as_prover9(sentence.antecedent, config, depth + 1)} -> {as_prover9(sentence.consequent, config, depth + 1)})"

    elif isinstance(sentence, Implied):
        return f"({as_prover9(sentence.consequent, config, depth + 1)} <- {as_prover9(sentence.antecedent, config, depth + 1)})"

    elif isinstance(sentence, Iff):
        return f"({as_prover9(sentence.left, config, depth + 1)} <-> {as_prover9(sentence.right, config, depth + 1)})"


    elif isinstance(sentence, Term):
        predicate = format_predicate(sentence.predicate)
        if not sentence.bindings:
            return predicate
        args = ", ".join(
            format_var(v) if isinstance(v, Variable) else format_value(v) for v in sentence.bindings.values())
        return f"{predicate}({args})"

    else:
        raise ValueError(f"Unsupported sentence type: {type(sentence)}")


def prover9_problem(theory: Theory, conjecture: Optional[Sentence] = None) -> str:
    """
    Generate a complete Prover9 problem from a theory and an optional conjecture.

    :param theory: The theory containing axioms
    :param conjecture: An optional conjecture to prove
    :return: Prover9 representation of the problem

    Example:
    >>> from typedlogic import Theory, PredicateDefinition, Term, Variable, Forall, Implies
    >>> X = Variable("X", "str")
    >>> Y = Variable("Y", "str")
    >>> theory = Theory(
    ...     name="example",
    ...     predicate_definitions=[
    ...         PredicateDefinition("P", {"x": "str"}),
    ...         PredicateDefinition("Q", {"x": "str", "y": "str"}),
    ...     ],
    ...     sentence_groups=[
    ...         SentenceGroup("axioms", sentences=[
    ...             Forall([X], Implies(Term("P", X), Term("Q", X, Y)))
    ...         ])
    ...     ]
    ... )
    >>> conjecture = Forall([X, Y], Implies(Term("P", X), Term("Q", X, Y)))
    >>> print(prover9_problem(theory, conjecture))
    formulas(assumptions).
        all x ((P(x) -> Q(x, y))).
    end_of_list.
    <BLANKLINE>
    formulas(goals).
        all x y ((P(x) -> Q(x, y))).
    end_of_list.
    """
    lines = []

    # Assumptions (axioms)
    lines.append("formulas(assumptions).")
    for sentence in theory.sentences:
        lines.append(f"    {as_prover9(sentence)}.")
    lines.append("end_of_list.")
    lines.append("")

    # Goals (conjecture)
    if conjecture:
        lines.append("formulas(goals).")
        lines.append(f"    {as_prover9(conjecture)}.")
        lines.append("end_of_list.")

    return "\n".join(lines)


def transform_sentence_chained(sentence: Sentence, rules: Iterable[Callable[[Sentence], Sentence]]) -> Sentence:
    """
    Transform a sentence using a chain of rules.

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> A = Term("A")
        >>> B = Term("B")
        >>> C = Term("C")
        >>> D = Term("D")
        >>> sentence = A | (B & C)
        >>> negate_conjunctions = lambda s: Or(*[~op for op in s.operands]) if isinstance(s, And) else s
        >>> reduce_singletons = lambda s: s.operands[0] if isinstance(s, And) and len(s.operands) == 1 else s
        >>> transform_sentence_chained(sentence, [negate_conjunctions, reduce_singletons])
        Or(A, Or(Not(B), Not(C)))

    :param sentence:
    :param rules:
    :return:
    """
    for rule in rules:
        sentence = transform_sentence(sentence, rule)
    return sentence

def transform_sentence(sentence: Sentence, rule: Callable[[Sentence], Sentence], varmap: Optional[Dict[str, Variable]] = None) -> Sentence:
    """
    Transform a sentence recursively using a rule.

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> A = Term("A")
        >>> B = Term("B")
        >>> C = Term("C")
        >>> D = Term("D")
        >>> sentence = A | (B & (C | ~D))
        >>> assert transform_sentence(sentence, lambda s: s) == sentence

        >>> def negate_conjunctions(sentence: Sentence) -> Optional[Sentence]:
        ...     if isinstance(sentence, And):
        ...         return Or(*[~op for op in sentence.operands])
        ...     return None
        >>> transform_sentence(B & C, negate_conjunctions)
        Or(Not(B), Not(C))
        >>> transform_sentence(A | (B & C), negate_conjunctions)
        Or(A, Or(Not(B), Not(C)))

    :param sentence:
    :param rule:
    :param varmap:
    :return:
    """
    varmap = varmap or {}
    new_sentence = rule(sentence)
    if new_sentence:
        if sentence != new_sentence:
            sentence = transform_sentence(new_sentence, rule, varmap)
        else:
            sentence = new_sentence
    if isinstance(sentence, QuantifiedSentence):
        return type(sentence)(sentence.variables, transform_sentence(sentence.sentence, rule, varmap))
    elif isinstance(sentence, BooleanSentence):
        return type(sentence)(*[transform_sentence(op, rule, varmap) for op in sentence.operands])
    elif isinstance(sentence, Term):
        return sentence
    elif isinstance(sentence, Extension):
        return sentence.to_model_object()
    else:
        raise ValueError(f"Unknown sentence type {type(sentence)} // {sentence}")


def replace_constants(sentence: Sentence, constant_map: Dict[str, Any] ) -> Sentence:
    """
    Replace constants in a sentence with new values.

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> Z = Variable("Z", "str")
        >>> A = Term("A", X)
        >>> B = Term("B", Y)
        >>> C = Term("C", Z)
        >>> constant_map = {"X": "foo", "Y": "bar", "Z": "baz"}
        >>> replace_constants(A & B, constant_map)
        And(A(foo), B(bar))


    :param sentence:
    :param constant_map:
    :return:
    """
    if isinstance(sentence, QuantifiedSentence):
        return type(sentence)(sentence.variables, replace_constants(sentence.sentence, constant_map))
    elif isinstance(sentence, BooleanSentence):
        return type(sentence)(*[replace_constants(op, constant_map) for op in sentence.operands])
    elif isinstance(sentence, Term):
        def _repl(v: Any) -> Any:
            if isinstance(v, Variable):
                if v.name in constant_map:
                    return constant_map[v.name]
            return v
        return Term(sentence.predicate, {k: _repl(v) for k, v in sentence.bindings.items()})
    else:
        raise ValueError(f"Unknown sentence type {sentence}")


def reduce_singleton(sentence: Sentence) -> Sentence:
    """
    Reduce singleton conjunctions and disjunctions in a sentence.

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> A = Term("A")
        >>> B = Term("B")
        >>> C = Term("C")
        >>> D = Term("D")
        >>> sentence = And(A, B)
        >>> reduce_singleton(sentence)
        And(A, B)
        >>> reduce_singleton(And(A))
        A
        >>> reduce_singleton(Or(A))
        A
        >>> transform_sentence(And(A, Or(B), C), reduce_singleton)
        And(A, B, C)
        >>> transform_sentence(And(And(A)), reduce_singleton)
        A
        >>> transform_sentence(And(A, Or(Or(B)), And(Or(And(C)))), reduce_singleton)
        And(A, B, C)

    :param sentence:
    :return:
    """
    if isinstance(sentence, And) and len(sentence.operands) == 1:
        return sentence.operands[0]
    if isinstance(sentence, Or) and len(sentence.operands) == 1:
        return sentence.operands[0]
    return sentence

def simplify(sentence: Sentence) -> Sentence:
    """
    Simplify a sentence by reducing nested conjunctions and disjunctions.

    >>> from typedlogic import And, Or, Variable, Term, Forall
    >>> A = Term("A")
    >>> B = Term("B")
    >>> C = Term("C")
    >>> D = Term("D")
    >>> simplify(And(A, And(B, And(C, D))))
    And(A, B, C, D)
    >>> simplify(And(And(A)))
    A

    :param sentence:
    :return:
    """
    if isinstance(sentence, (And, Or)):
        operands = [simplify(op) for op in sentence.operands]
        if len(operands) == 1:
            return simplify(operands[0])
        op_type = type(sentence)
        new_operands: List[Sentence] = []
        for op in operands:
            if isinstance(op, op_type):
                if not isinstance(op, (And, Or)):
                    raise AssertionError
                new_operands.extend(op.operands)
            else:
                new_operands.append(op)
        sentence = op_type(*new_operands)
    if isinstance(sentence, (Exists, Forall)):
        typ = type(sentence)
        qs = sentence.sentence
        if typ == type(qs):
            if not isinstance(qs, (Exists, Forall)):
                raise AssertionError
            return typ(sentence.variables + qs.variables, simplify(qs.sentence))
    if isinstance(sentence, Not):
        negated = simplify(sentence.negated)
        if isinstance(negated, Not):
            return negated.negated
        return Not(negated)
    return sentence

def distribute_and_over_or(sentence: Sentence) -> Sentence:
    """
    Distribute AND over OR in a sentence.

    A ∨ (B1 ∧ B2 ∧ ... ∧ Bn) ≡ (A ∨ B1) ∧ (A ∨ B2) ∧ ... ∧ (A ∨ Bn)

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> A = Term("A")
        >>> B = Term("B")
        >>> C = Term("C")
        >>> D = Term("D")
        >>> distribute_and_over_or(A | (B & C))
        And(Or(B, A), Or(C, A))
        >>> distribute_and_over_or((A & B) | (C & D))
        And(Or(C, A), Or(D, A), Or(C, B), Or(D, B))

        >>> distribute_and_over_or(A | And(B, C, D))
        And(Or(B, A), Or(C, A), Or(D, A))

        >>> distribute_and_over_or(A | B)
        Or(A, B)

    :param sentence:
    :return:
    """
    if not isinstance(sentence, Or):
        return sentence
    sentence = _distribute_sentence(sentence, And, Or)
    # sentence = transform_sentence(sentence, reduce_singleton)
    return sentence

def _distribute_sentence(sentence: Sentence, op1: Type[BooleanSentence], op2: Type[BooleanSentence]) -> Sentence:
    # adapted from sympy
    dfunc = lambda t: _distribute_sentence(*t)
    sentence = simplify(sentence)

    if isinstance(sentence, op2):
        for arg in sentence.operands:
            if isinstance(arg, op1):
                conj = arg
                break
        else:
            return sentence
        rest = op2(*[a for a in sentence.operands if a is not conj])
        tups = [(op2(c, rest), op1, op2) for c in conj.operands]
        if not tups:
            raise ValueError("Expected at least one operand")
        mapped = list(map(dfunc, tups))
        return simplify(op1(*mapped))
    elif isinstance(sentence, op1):
        tups = [(x, op1, op2) for x in sentence.operands]
        if not tups:
            raise ValueError("Expected at least one operand")
        mapped = list(map(dfunc, tups))
        if len(mapped) == 1:
            return mapped[0]
        else:
            return simplify(op1(*mapped))
    else:
        return sentence

def flatten_nested_conjunctions_and_disjunctions(sentence: Sentence) -> Sentence:
    """
    Flatten nested conjunctions and disjunctions in a sentence.

    Replace (A ∧ B) ∧ C with A ∧ B ∧ C
    Replace (A ∨ B) ∨ C with A ∨ B ∨ C

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> Z = Variable("Z", "str")
        >>> sentence = And(And(Term("Q", X), Term("R", Y)), Term("S", Z))
        >>> flatten_nested_conjunctions_and_disjunctions(sentence)
        And(Q(?X), R(?Y), S(?Z))

    Note: does not recurse, use :ref:`transform_sentence` to apply to all sub-sentences.

    :param sentence:
    :return:
    """
    if isinstance(sentence, (And, Or)):
        new_ops = []
        for op in sentence.operands:
            if isinstance(op, type(sentence)):
                new_ops.extend(op.operands)
            else:
                new_ops.append(op)
        return type(sentence)(*new_ops)
    return sentence



def skolemize(sentence: Sentence, universal_vars: Optional[List[Variable]] = None, substitution_map: Optional[Dict[str, Term]] = None) -> Sentence:
    """
    Skolemize a sentence.

    Examples:

        >>> from typedlogic import And, Or, Variable, Term, Forall, Exists
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> Z = Variable("Z", "str")
        >>> # no universal quantifiers
        >>> skolemize(Exists([X], Term("P", X)))
        P(sk__1)
        >>> skolemize(Forall([X], Exists([Y], Term("P", X, Y))))
        Forall([X: str] : P(?X, sk__1(?X)))
        >>> skolemize(Forall([X], Exists([Y, Z], Term("P", X, Y, Z))))
        Forall([X: str] : P(?X, sk__1(?X), sk__2(?X)))
        >>> skolemize(Forall([X], And(Exists([Y, Z], Term("P", X, Y, Z)), Exists([Y], Term("Q", X, Y)))))
        Forall([X: str] : And(P(?X, sk__1(?X), sk__2(?X)), Q(?X, sk__3(?X))))


    :param sentence:
    :param universal_vars:
    :param substitution_map:
    :return:
    """
    if substitution_map is None:
        substitution_map = {}
    if not universal_vars:
        universal_vars = []
    if isinstance(sentence, Forall):
        return Forall(sentence.variables, skolemize(sentence.sentence, universal_vars + sentence.variables, substitution_map))
    if isinstance(sentence, Exists):
        vars_to_skolemize = [v for v in sentence.variables if v not in universal_vars]
        n = len(substitution_map)
        for v in vars_to_skolemize:
            n += 1
            skolem_term = Term(f"sk__{n}", *universal_vars)
            substitution_map[v.name] = skolem_term
        return skolemize(sentence.sentence, universal_vars, substitution_map)
    if isinstance(sentence, BooleanSentence):
        return type(sentence)(*[skolemize(op, universal_vars, substitution_map) for op in sentence.operands])
    elif isinstance(sentence, Term):
        return Term(sentence.predicate,
                    {
                        k: substitution_map[v.name] if isinstance(v, Variable) and v.name in substitution_map else v
                        for k, v in sentence.bindings.items()
                    })
    else:
        raise ValueError(f"Unknown sentence type {type(sentence)} // {sentence}")


def to_cnf(sentence: Sentence, skip_skolemization=False) -> Sentence:
    """
    Convert a sentence to conjunctive normal form.

    Examples:

        >>> from typedlogic.profiles import SortedLogic
        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> Z = Variable("Z", "str")
        >>> P = Term("P")
        >>> Q = Term("Q")
        >>> R = Term("R")
        >>> S = Term("S")
        >>> T = Term("T")
        >>> to_cnf(P)
        P
        >>> to_cnf(P & Q)
        And(P, Q)
        >>> to_cnf(P | Q)
        Or(P, Q)
        >>> to_cnf(P >> Q)
        Or(Not(P), Q)
        >>> to_cnf(~(P | Q) | R)
        And(Or(Not(P), R), Or(Not(Q), R))
        >>> to_cnf((P & Q) >> R)
        Or(Not(P), Not(Q), R)
        >>> to_cnf(((P & Q) | (R & S)))
        And(Or(R, P), Or(S, P), Or(R, Q), Or(S, Q))
        >>> to_cnf((P | Q) & (R | S))
        And(Or(P, Q), Or(R, S))
        >>> sentence = Or(And(Term("Q", X), Term("R", Y)), Term("S", Z))
        >>> to_cnf(sentence)
        And(Or(Q(?X), S(?Z)), Or(R(?Y), S(?Z)))
        >>> to_cnf(P >> (Q | R))
        Or(Not(P), Q, R)

    :param sentence:
    :param skip_skolemization:
    :return:
    """
    # Eliminate XORs
    sentence = transform_sentence_chained(sentence, [expand_xor, expand_exactly_one])
    # Eliminate implications and equivalences
    sentence = eliminate_all_implications_recursive(sentence)
    # Move negations inward
    sentence = transform_sentence_chained(sentence, [apply_demorgans, apply_quantifier_negation])
    # Standardize variables - TODO
    # Skolemize
    if not skip_skolemization:
        sentence = skolemize(sentence)
    #def raise_if_exists(s: Sentence) -> Sentence:
    #    if isinstance(s, Exists):
    #        raise NotInProfileError("Exists not allowed in CNF")
    #    return s

    #transform_sentence(sentence, raise_if_exists)
    # Drop universal quantifiers
    sentence = transform_sentence(sentence, lambda s: s.sentence if isinstance(s, Forall) else s)
    # Distribute OR over AND
    sentence = transform_sentence(sentence, distribute_and_over_or)
    return sentence

def to_cnf_lol(sentence: Sentence, **kwargs) -> List[List[Sentence]]:
    """
    Convert a sentence to a list of lists of sentences in conjunctive normal form.

    Examples:

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> Z = Variable("Z", "str")
        >>> sentence = Or(And(Term("Q", X), Term("R", Y)), Term("S", Z))
        >>> to_cnf_lol(sentence)
        [[Q(?X), S(?Z)], [R(?Y), S(?Z)]]

        >>> to_cnf_lol(Term("P"))
        [[P]]

        >>> to_cnf_lol(Or(Term("P"), Term("Q")))
        [[P, Q]]
        >>> to_cnf_lol(And(Term("P"), Term("Q")))
        [[P], [Q]]

    :param sentence:
    :param kwargs:
    :return:
    """
    sentence = to_cnf(sentence, **kwargs)
    sentence = simplify(sentence)
    if not isinstance(sentence, And):
        sentence = And(sentence)
    return [
        list(op.operands) if isinstance(op, Or) else [op] for op in sentence.operands
    ]

def to_horn_rules(sentence: Sentence, allow_disjunctions_in_head=False, allow_goal_clauses=None) -> List[Sentence]:
    """
    Convert a sentence to a list of Horn rules.

    Examples:

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> Z = Variable("Z", "str")
        >>> P = Term("P")
        >>> Q = Term("Q")
        >>> R = Term("R")
        >>> S = Term("S")
        >>> print(as_prolog(to_horn_rules(R >> (Q & P))))
        q :- r.
        p :- r.
        >>> print(as_prolog(to_horn_rules((R & S) >> P)))
        p :- r, s.

    :param sentence:
    :param allow_disjunctions_in_head:
    :return:
    """
    if allow_goal_clauses is None:
        allow_goal_clauses = allow_disjunctions_in_head
    sentence = transform_sentence(sentence, lambda s: s.to_model_object() if isinstance(s, Extension) else s)
    sentence = simplify(sentence)
    # TODO: check if already in horn profile
    cnf_lol = to_cnf_lol(sentence, skip_skolemization=True)
    rules: List[Sentence] = []
    for dnf_sentence in cnf_lol:
        # separate into positive and negative literals
        positive = [] # head
        negative = [] # body
        for lit in dnf_sentence:
            if isinstance(lit, Not):
                negative.append(lit.negated)
            else:
                positive.append(lit)
        if not positive and not negative:
            # The empty clause, consisting of no literals (which is equivalent to false) is a goal clause
            rules.append(Or())
            continue
        # a horn clause is a disjunction of literals with at most one positive literal.
        if len(positive) > 1 and not allow_disjunctions_in_head:
            # not a horn clause (and not disjunctive datalog).
            # we could potentially generate multiple rules here, but this could
            # lead to stratified negation issues. We choose the last to be order preserving
            pos = positive[-1]
            other_pos = positive[:-1]
            anded = negative + [Not(other) for other in other_pos]
            rules.append(Implies(And(*anded), pos))
            # TODO: uncomment this to generate multiple rules
            #for pos in positive:
            #    other_pos = [p for p in positive if p != pos]
            #    anded = negative + [Not(other) for other in other_pos]
            #    rules.append(Implies(And(*anded), pos))
            continue
        # a unit clause is a disjunction of literals with no negative literals
        body = And(*negative) if len(negative) != 1 else negative[0]
        if not positive and allow_goal_clauses:
            # A Horn clause without a positive literal is a goal clause.
            # Or() == False
            rules.append(Implies(body, Or()))
        if len(positive) == 1:
            head = positive[0]
            rules.append(Implies(body, head))
        if len(positive) > 1:
            # we must be in disjunctive datalog at this pont
            head = Or(*positive)
            rules.append(Implies(body, head))
    return rules

def expand_xor(sentence: Sentence) -> Sentence:
    """
    Expand XOR in a sentence.

    Replace A ⊕ B with (A ∨ B) ∧ ¬(A ∧ B)

        >>> from typedlogic import And, Or, Variable, Term, Forall
        >>> P = Term("P")
        >>> Q = Term("Q")
        >>> expand_xor(P ^ Q)
        And(Or(P, Q), Not(And(P, Q)))

    :param sentence:
    :return:
    """
    if not isinstance(sentence, Xor):
        return sentence
    # expand XOR to an OR plus an AND, where len(operands) may be > 2
    operands = sentence.operands
    if len(operands) == 1:
        return operands[0]
    return And(Or(*operands), Not(And(*operands)))


def expand_exactly_one(sentence: Sentence) -> Sentence:
    if not isinstance(sentence, ExactlyOne):
        return sentence
    # expand XOR to an OR plus an AND, where len(operands) may be > 2
    operands = sentence.operands
    if len(operands) == 1:
        return operands[0]
    if len(operands) == 2:
        return And(Or(*operands), Not(And(*operands)))
    return Or(
        [And(op,
             Not(Or([op2 for op2 in operands if op2 != op])))
         for op in operands])


def eliminate_all_implications_recursive(sentence: Sentence) -> Sentence:
    """
    Replace A → B with ¬A ∨ B
    Replace A ↔ B with (A → B) ∧ (B → A), then apply the above rule

    Replace A ↔ B with (A → B) ∧ (B → A)

        >>> from typedlogic import Iff, Variable, Term, Forall
        >>> P = Term("P")
        >>> Q = Term("Q")
        >>> R = Term("R")
        >>> print(eliminate_all_implications_recursive(P >> Q))
        (~P) | (Q)
        >>> print(eliminate_all_implications_recursive(Iff(P, Q)))
        ((~P) | (Q)) & ((~Q) | (P))
        >>> print(eliminate_all_implications_recursive(P & ( Q >> R)))
        (P) & ((~Q) | (R))

    :param sentence:
    :return:
    """

    def eliminate_implies(sentence: Sentence) -> Sentence:
        if isinstance(sentence, Implies):
            return Or(Not(sentence.antecedent), sentence.consequent)
        return sentence

    def eliminate_implied(sentence: Sentence) -> Sentence:
        if isinstance(sentence, Implied):
            return Implies(sentence.antecedent, sentence.consequent)
        return sentence

    def eliminate_iff(sentence: Sentence) -> Sentence:
        if isinstance(sentence, Iff):
            return And(
                Implies(sentence.left, sentence.right),
                Implies(sentence.right, sentence.left),
            )
        return sentence

    return transform_sentence_chained(sentence, [eliminate_iff, eliminate_implied, eliminate_implies])


def eliminate_iff(sentence: Sentence) -> Sentence:
    """
    Eliminate iff from a sentence.

    Replace A ↔ B with (A → B) ∧ (B → A)

        >>> from typedlogic import Iff, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> sentence = Iff(Term("Q", X), Term("R", Y))
        >>> eliminate_iff(sentence)
        And(Implies(Q(?X), R(?Y)), Implies(R(?Y), Q(?X)))

        >>> sentence = Forall([X], Iff(Term("Q", X), Term("R", Y)))
        >>> eliminate_iff(sentence)
        Forall([X: str] : Iff(Q(?X), R(?Y)))

    Note: does not recurse, use :ref:`transform_sentence` to apply to all sub-sentences.

    :param sentence:
    :return:
    """
    if not isinstance(sentence, Iff):
        return sentence
    return And(
        Implies(sentence.left, sentence.right),
        Implies(sentence.right, sentence.left),
    )

def eliminate_implied(sentence: Sentence) -> Sentence:
    """
    Eliminate implied from a sentence.

    This simply reverses the direction of the implication.

        >>> from typedlogic import Implied, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> sentence = Implied(Term("Q", X), Term("R", Y))
        >>> eliminate_implied(sentence)
        Implies(R(?Y), Q(?X))

    Note: does not recurse, use :ref:`transform_sentence` to apply to all sub-sentences.

    :param sentence:
    :return:
    """
    if not isinstance(sentence, Implied):
        return sentence
    return Implies(sentence.operands[1], sentence.operands[0])

def eliminate_implies(sentence: Sentence) -> Sentence:
    """
    Eliminate implies from a sentence in translation to CNF

    Replace A → B with ¬A ∨ B

        >>> from typedlogic import Implies, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> sentence = Implies(Term("Q", X), Term("R", Y))
        >>> eliminate_implies(sentence)
        Or(Not(Q(?X)), R(?Y))

    Note: does not recurse, use :ref:`transform_sentence` to apply to all sub-sentences.
    """
    if not isinstance(sentence, Implies):
        return sentence
    return Or(Not(sentence.operands[0]), sentence.operands[1])




def apply_demorgans(sentence: Sentence) -> Sentence:
    """
    Apply De Morgan's laws to a sentence.

    Replace ¬(A ∧ B) with ¬A ∨ ¬B
    Replace ¬(A ∨ B) with ¬A ∧ ¬B

        >>> from typedlogic import And, Or, Not, Variable, Term, Forall
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> sentence = Not(And(Term("Q", X), Term("R", Y)))
        >>> apply_demorgans(sentence)
        Or(Not(Q(?X)), Not(R(?Y)))

    Note: does not recurse, use :ref:`transform_sentence` to apply to all sub-sentences.

    :param sentence:
    :return:
    """
    if not isinstance(sentence, Not):
        return sentence
    negated = sentence.negated
    if isinstance(negated, And):
        return Or(*[Not(op) for op in negated.operands])
    if isinstance(negated, Or):
        return And(*[Not(op) for op in negated.operands])
    return sentence

def apply_quantifier_negation(sentence: Sentence) -> Sentence:
    """
    Apply negation of quantifiers to a sentence.

    Replace ¬∀x P(x) with ∃x ¬P(x)
    Replace ¬∃x P(x) with ∀x ¬P(x)

        >>> from typedlogic import Forall, Variable, Term
        >>> X = Variable("X", "str")
        >>> sentence = Not(Forall([X], Term("P", X)))
        >>> apply_quantifier_negation(sentence)
        Exists(X: str : Not(P(?X)))

    :param sentence:
    :return:
    """
    if not isinstance(sentence, Not):
        return sentence
    negated = sentence.negated
    if isinstance(negated, Forall):
        return Exists(negated.variables, Not(negated.sentence))
    if isinstance(negated, Exists):
        return Forall(negated.variables, Not(negated.sentence))
    return sentence

def force_stratification(horn_rules: List[Sentence]) -> List[Sentence]:
    """
    Force stratification of a list of horn rules.

    If the program is not stratified, remove a rule using negation that causes the issue
    and recurse.

    Note that this will weaken the program, but it will be stratified.

        >>> from typedlogic import Implies, Variable, Term, Forall
        >>> P = Term("P")
        >>> Q = Term("Q")
        >>> R = Term("R")
        >>> sentence = Iff(P&Q, R)
        >>> rules = to_horn_rules(sentence)
        >>> force_stratification(rules)
        [Implies(And(P, Q), R), Implies(R, P), Implies(R, Q)]
        >>> sentence = And(Iff(P|Q, R), Iff(P, ~Q))
        >>> rules = to_horn_rules(sentence)
        >>> rules
        [Implies(P, R), Implies(Q, R), Implies(And(R, Not(P)), Q), Implies(And(Not(Q)), P)]
        >>> force_stratification(rules)
        [Implies(P, R), Implies(Q, R), Implies(And(R, Not(P)), Q)]

    :param horn_rules:
    :return:
    """
    pmap: Dict[str, List[Tuple[str, bool]]] = defaultdict(list)
    edge_to_rules = defaultdict(list)
    for i, rule in enumerate(horn_rules):
        if isinstance(rule, Implies):
            head = rule.consequent
            if not isinstance(head, Term):
                continue
            pred = head.predicate
            body = rule.antecedent
            if isinstance(body, And):
                terms = list(body.operands)
            else:
                terms = [body]
            for term in terms:
                negated = False
                if isinstance(term, Not):
                    term = term.negated
                    negated = True
                if not isinstance(term, Term):
                    continue
                pmap[pred].append((term.predicate, negated))
                if negated:
                    edge_to_rules[(pred, term.predicate)].append(i)
    is_stratified, edge, _ = analyze_datalog_program(list(pmap.items()))
    if not is_stratified:
        if not edge:
            raise AssertionError
        candidates = edge_to_rules[edge] if edge in edge_to_rules else []
        if not candidates:
            raise AssertionError(f"Stratification failed; cannot find {edge} in {edge_to_rules}")
        rule_to_remove = candidates[0]
        horn_rules = [rule for i, rule in enumerate(horn_rules) if i != rule_to_remove]
        return force_stratification(horn_rules)
    return horn_rules

def ensure_terms_positional(theory: Theory):
    """
    Ensure that all terms in a theory have all positions filled and ordered.

        >>> from typedlogic import Term, Variable, PredicateDefinition, Theory
        >>> X = Variable("X", "str")
        >>> Y = Variable("Y", "str")
        >>> theory = Theory(predicate_definitions=[PredicateDefinition("P", {"x": "str", "y": "str"})])
        >>> s1 = Term("P", X, Y)
        >>> s2 = Term("P", {"x": X, "y": Y})
        >>> s3 = Term("P", {"y": Y, "x": X})
        >>> s4 = Term("P", {"y": Y})
        >>> s5 = Term("P", {})
        >>> theory.add(s1)
        >>> theory.add(s2)
        >>> theory.add(s3)
        >>> theory.add(s4)
        >>> theory.add(s5)
        >>> ensure_terms_positional(theory)
        >>> for s in theory.sentences:
        ...     print(s)
        P(?X, ?Y)
        P(?X, ?Y)
        P(?X, ?Y)
        P(None, ?Y)
        P(None, None)

    :param theory:
    :return:
    """
    def tr(s: Sentence):
        if isinstance(s, Term):
            pds = [pd for pd in theory.predicate_definitions if pd.predicate == s.predicate]
            if len(pds) != 1:
                # could include builtins
                return
            pd = pds[0]
            if not s.positional:
                s.bindings = {k: s.bindings.get(k) for k, v in pd.arguments.items()}
    for s in theory.sentences:
        transform_sentence(s, tr)








import logging
import subprocess
import tempfile
from dataclasses import dataclass, field
from typing import ClassVar, Iterator, List, Optional

from typedlogic import Sentence
from typedlogic.compilers.prover9_compiler import Prover9Compiler
from typedlogic.profiles import MixedProfile, OpenWorld, Profile, Unrestricted
from typedlogic.solver import Model, Solution, Solver

logger = logging.getLogger(__name__)

@dataclass
class Prover9Solver(Solver):
    """
    A solver that uses Prover9.

    Prover9 is an automated theorem prover for first-order and equational logic

    See [The Prover9 site](https://www.cs.unm.edu/~mccune/prover9/)

    Note that in order to use this integration, you need to [install Prover9](https://www.cs.unm.edu/~mccune/prover9/)
    and have it on your path.

    Example:
    -------
        >>> from typedlogic.integrations.solvers.prover9 import Prover9Solver
        >>> from typedlogic.parsers.pyparser import PythonParser
        >>> import tests.theorems.simple_contradiction as sc
        >>> solver = Prover9Solver()
        >>> solver.load(sc)
        >>> r = solver.check()
        >>> r.satisfiable
        False


    This solver does implements the open-world assumption.

    """

    exec_name: str = field(default="prover9")
    profile: ClassVar[Profile] = MixedProfile(Unrestricted(), OpenWorld())

    def models(self) -> Iterator[Model]:
        r = self.check()
        if r.satisfiable:
            yield Model()

    def prove(self, sentence: Sentence) -> Optional[bool]:
        proved = self._run(goals=[sentence])
        return proved

    def _run(self, goals: Optional[List[Sentence]] = None) -> bool:
        compiler = Prover9Compiler()
        #print(f"THEORY; n_sentences: {len(self.base_theory.sentences)}")
        program = compiler.compile(self.base_theory, goals=goals)

        #print(program)

        with tempfile.NamedTemporaryFile(suffix=".prover9", mode="w") as fp:
            fp.write(program)
            fp.flush()
            res = subprocess.run([self.exec_name, "-f", fp.name], capture_output=True)
            if res.returncode not in (0, 2):
                logger.error(res.stdout.decode())
                raise ValueError(f"Prover9 failed with return code {res.returncode}")

            if "THEOREM PROVED" in res.stdout.decode():
                return True
            else:
                return False

    def check(self) -> Solution:
        proved = self._run()
        unsat = not self.base_theory.goals and not proved
        return Solution(satisfiable=unsat)

    def dump(self) -> str:
        compiler = Prover9Compiler()
        return compiler.compile(self.base_theory)
